#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _phxtof_calibration(energy, alpha, beta, gamma):
    r"""Pulse Height x Time Of Flight correction model from EPD Data Product
    Guide"""
    return 1 / (0.5 * (1 + alpha * (np.tanh((energy - beta) / gamma) + 1)))


def _extof_calibration(energy, alpha, beta, gamma):
    r"""Energy x Time Of Flight correction model from EPD Data Product Guide"""
    return 1 / (0.5 * (1 + alpha * (1 - np.tanh((energy - beta) / gamma) + 1)))


def eis_proton_correction(flux_eis):
    r"""Corrects proton flux values based on FPI/HPCA/EPD-EIS cross
    calibration. Correction to the EIS PHxTOF data are made by applying an
    energy-dependent numerical correction of the form:

    .. math:

        E_{PHxTOF} = \frac{1}{0.5*\left [1 + \alpha_{PH} \left (
        \operatorname{tanh}\left ( \frac{E - \beta_{PH}}{\gamma_{PH}} + 1
        \right ) \right ) \right ]}


    where E is energy and  :math:`\alpha_{PH} = -0.3` , :math:`\beta_{PH} =
    0.049` , :math:`\gamma_{PH} = 0.001` are coefficients. Minor adjustments
    were also made to the lowest energy EIS ExTOF data to correct for foil
    efficiencies. This correction is of the form:

    .. math:

        E_{ExTOF} = \frac{1}{0.5*\left [1 + \alpha_{E} \left (1 -
        \operatorname{tanh}\left ( \frac{E - \beta_{E}}{\gamma_{E}} + 1
        \right ) \right ) \right ]}

    where E is energy and :math:`\alpha_{E} = -0.3` , :math:`\beta_{E} =
    0.049` , :math:`\gamma_{E} = 0.001` are coefficients.

    Parameters
    ----------
    flux_eis : xarray.DataArray
        Omni-directional energy spectrum from EPD-EIS.

    Returns
    -------
    flux_eis_corr : xarray.DataArray
        Cross-calibrated omni-directional energy spectrum from EIS-EPD.

    See Also
    --------
    pyrfu.mms.get_eis_allt, pyrfu.mms.eis_omni

    """

    assert isinstance(flux_eis, (xr.Dataset, xr.DataArray)), "flux_eis must be a xarray"

    #  Coefficients from EPD Data Product Guide
    alpha_, beta_, gamma_ = [-0.3, 49e-3, 1e-3]

    # Pulse Height x Time Of Flight (PHxTOF) energy correction factor
    energy_phxtof = flux_eis.energy.data[:7]
    phxtof_corr = _phxtof_calibration(energy_phxtof, alpha_, beta_, gamma_)

    # Energy x Time Of Flight (ExTOF) energy correction factor
    energy_extof = flux_eis.energy.data[7:]
    extof_corr = _extof_calibration(energy_extof, alpha_, beta_, gamma_)

    eis_corr = np.hstack([phxtof_corr, extof_corr])

    if isinstance(flux_eis, xr.Dataset):
        scopes_eis = list(filter(lambda x: x[0] == "t", flux_eis))
        out_keys = list(filter(lambda x: x not in scopes_eis, flux_eis))
        out_dict = {k: flux_eis[k] for k in out_keys}

        for scope in scopes_eis:
            out_dict[scope] = flux_eis[scope].copy()
            out_dict[scope].data *= eis_corr

        flux_eis_corr = xr.Dataset(out_dict)
    else:
        # Apply correction to omni-directional energy spectrum
        flux_eis_corr = flux_eis.copy()
        flux_eis_corr.data *= eis_corr

    return flux_eis_corr



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def ts_vec_xyz(time, data, attrs: dict = None):
    r"""Create a time series containing a 1st order tensor.

    Parameters
    ----------
    time : ndarray
        Array of times.
    data : ndarray
        Data corresponding to the time list.
    attrs : dict, Optional
        Attributes of the data list.

    Returns
    -------
    out : xarray.DataArray
        1st order tensor time series.

    """

    # Check input type
    assert isinstance(time, np.ndarray), "time must be a numpy.ndarray"
    assert isinstance(data, np.ndarray), "data must be a numpy.ndarray"

    # Check input shape must be (n, 3)
    assert data.ndim == 2 and data.shape[1] == 3
    assert len(time) == len(data), "Time and data must have the same length"

    if attrs is None or not isinstance(attrs, dict):
        attrs = {}

    out = xr.DataArray(
        data,
        coords=[time[:], ["x", "y", "z"]],
        dims=["time", "comp"],
        attrs=attrs,
    )

    out.attrs["TENSOR_ORDER"] = 1

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from scipy.optimize import curve_fit

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def estimate_phase_speed(f_k_power, freq, k, f_min: float = 100.0):
    r"""Simple function to estimate the phase speed from the frequency
    wave number power spectrum. Fits :math:`f = v k/ 2 \\pi` to the
    power spectrum.

    Parameters
    ----------
    f_k_power : ndarray
        2D array of powers.
    freq : ndarray
        1D array of frequencies.
    k : ndarray
        1D array of wave numbers.
    f_min : float, Optional
        Set low frequency threshold of points used to estimate the speed.
        Default ``f_min`` = 100.

    Returns
    -------
    vph : float
        Estimated phase speed by fitting linear dispersion relation to
        data.

    Notes
    -----
    Draft version but seems to work well. Does not yet handle multiple
    modes in the same power spectrum.

    See also
    --------
    pyrfu.mms.fk_power_spectrum : Calculates the frequency-wave number
                                    power spectrum
    pyrfu.mms.probe_align_times : Returns times when f-a electrostatic
                                    waves can be characterized.

    """

    # Remove spurious points; specifically at large k.
    k_max = 2.0 * np.max(k) / 3.0
    power_temp = f_k_power
    rm_k = np.where(abs(k) > k_max)
    rm_f = np.where(freq < f_min)

    power_temp[:, rm_k] = 0.0
    power_temp[rm_f, :] = 0.0
    power_temp[np.isnan(power_temp)] = 0.0

    # Find position of maximum power to guess vph
    max_pos = np.unravel_index(np.argmax(power_temp), power_temp.shape)

    k_mat, f_mat = np.meshgrid(k, freq)
    max_f, max_k = [f_mat[max_pos], k_mat[max_pos]]

    # Initial guess
    vph_guess = max_f / max_k

    if vph_guess > 0.0:
        power_temp[:, k < 0.0] = 0
    else:
        power_temp[:, k > 0.0] = 0

    vph_range = [vph_guess / 3, vph_guess * 3]

    # Find all points around this guess vph
    highppos = np.where(power_temp > 0.3 * np.max(power_temp))

    p_power = power_temp[highppos]
    f_power = f_mat[highppos]
    k_power = k_mat[highppos]

    p_power2, f_power2, k_power2 = [[], [], []]

    for i, p_pow in enumerate(p_power):
        if (
            np.abs(vph_range[0])
            < np.abs(f_power[i] / k_power[i])
            < np.abs(vph_range[1])
        ):
            p_power2.append(p_pow)
            f_power2.append(f_power[i])
            k_power2.append(k_power[i])

    p_power2 = np.array(p_power2)
    f_power2 = np.array(f_power2)
    k_power2 = np.array(k_power2)

    weights = 1 + np.log10(p_power2 / np.max(p_power2))

    def fun(k_pow, f_pow):
        return f_pow * k_pow

    res = curve_fit(fun, k_power2, f_power2, p0=vph_guess, sigma=weights)
    vph = res[0][0] * 2 * np.pi

    return vph



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from scipy import constants

# Local imports
from ..pyrf.ts_tensor_xyz import ts_tensor_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def remove_imoms_background(n_i, v_gse_i, p_gse_i, n_bg_i, p_bg_i):
    r"""Remove the mode background population due to penetrating radiation
    from the the moments (density `n_i`, bulk velocity `v_gse_i` and
    pressure tensor `p_gse_i`) of the ion velocity distribution function
    using the method from [1]_.

    Parameters
    ----------
    n_i : xarray.DataArray
        Time series of the ion density.
    v_gse_i : xarray.DataArray
        Time series of the ion bulk velocity.
    p_gse_i : xarray.DataArray
        Time series of the ion pressure tensor.
    n_bg_i : xarray.DataArray
        Time series of the background ion number density.
    p_bg_i : xarray.DataArray
        Time series of the background ion pressure scalar.

    Returns
    -------
    n_i_new : xarray.DataArray
        Time series of the corrected ion number density.
    v_gse_i_new : xarray.DataArray
        Time series of the corrected ion bulk velocity.
    p_gse_i : xarray.DataArray
        Time series of the corrected ion pressure tensor.

    References
    ----------
    .. [1]  Gershman, D. J., Dorelli, J. C., Avanov,L. A., Gliese, U., Barrie,
            A., Schiff, C.,et al. (2019). Systematic uncertainties in plasma
            parameters reported by the fast plasma investigation on NASA's
            magnetospheric multiscale mission. Journal of Geophysical
            Research: Space Physics, 124, https://doi.org/10.1029/2019JA026980

    """

    m_p = constants.proton_mass

    # Correct the ion number density
    n_i_new = n_i - n_bg_i.data

    # Correct the ion bulk velocity
    v_gse_i_new = v_gse_i.copy()
    v_gse_i_new.data *= n_i.data[:, None] / n_i_new.data[:, None]

    # Correct the ion pressure tensor
    p_gse_i_new = np.zeros(p_gse_i.shape)
    n_old, v_old = [n_i.data, v_gse_i.data]
    n_new, v_new = [n_i_new.data, v_gse_i_new.data]

    for i, j in zip([0, 1, 2, 0, 0, 1], [0, 1, 2, 1, 2, 2]):
        p_gse_i_new[:, i, j] += p_gse_i.data[:, i, j]

        # TODO : use * instead of np.multiply??
        p_gse_i_new[:, i, j] += m_p * n_old * np.multiply(v_old[:, i], v_old[:, j])
        p_gse_i_new[:, i, j] -= m_p * n_new * np.multiply(v_new[:, i], v_new[:, j])

    # Remove isotropic background pressure
    p_bkg_mat = np.tile(np.eye(3, 3), (len(p_bg_i.data), 1, 1))
    p_bkg_mat *= p_bg_i.data[:, None, None]
    p_gse_i_new -= p_bkg_mat

    # Fill the lower left off diagonal terms using symetry of the
    # pressure tensor
    p_gse_i_new[:, 1, 0] = p_gse_i_new[:, 0, 1]
    p_gse_i_new[:, 2, 0] = p_gse_i_new[:, 0, 2]
    p_gse_i_new[:, 2, 1] = p_gse_i_new[:, 1, 2]

    # Create time series of the ion pressure tensor
    p_gse_i_new = ts_tensor_xyz(p_gse_i.time.data, p_gse_i_new)

    return n_i_new, v_gse_i_new, p_gse_i_new



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def iso86012datetime64(time):
    r"""Convert ISO8601 time format to datetime64 in ns units.

    Parameters
    ----------
    time : ndarray
        Time in ISO 8601 format

    Returns
    -------
    time_datetime64 : ndarray
        Time in datetime64 in ns units.

    See Also
    --------
    pyrfu.pyrf.datetime642iso8601

    """

    time_datetime64 = time.astype("datetime64[ns]")

    return time_datetime64



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import logging

# 3rd party imports
import numpy as np
import xarray as xr
from scipy import signal

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def psd(
    inp,
    n_fft: int = 256,
    n_overlap: int = 128,
    window: str = "hamming",
    d_flag: str = "constant",
    scaling: str = "density",
):
    r"""Estimate power spectral density using Welch's method.

    Welch's method [11]_ computes an estimate of the power spectral
    density by dividing the data into overlapping segments, computing a
    modified periodogram for each segment and averaging the
    periodograms.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of measurement values.
    n_fft : int, Optional
        Length of the FFT used, if a zero padded FFT is desired.
        Default to 256.
    n_overlap : int, Optional
        Number of points to overlap between segments. Default to 128.
    window : str, Optional
        Desired window to use. It is passed to `get_window` to generate
        the window values, which are DFT-even by default.
        See "get_window" or a list of windows and required parameters.
        Default Hanning
    d_flag : str, Optional
        Specifies how to detrend each segment. It is passed as the
        "type" argument to the"detrend" function. Default to "constant".
    scaling : str, Optional
        Selects between computing the power spectral density
        ('density') where `Pxx` has units of V**2/Hz and computing the
        power spectrum ("spectrum") where "Pxx" has units of V**2, if
        `x` is measured in V and "fs" is measured in Hz. Default to 'density'

    Returns
    -------
    out : xarray.DataArray
        Power spectral density or power spectrum of inp.

    References
    ----------
    .. [11] P. Welch, "The use of the fast Fourier transform for the
            estimation of power spectra: A method based on time
            averaging over short, modified periodograms",
            IEEE Trans. Audio Electroacoust. vol. 15, pp. 70-73, 1967.

    """

    if inp.ndim == 2 and inp.shape[-1] == 3:
        inp = np.abs(inp)

    if n_overlap is None:
        n_persegs = 256
        n_overlap = n_persegs / 2
    else:
        n_persegs = 2 * n_overlap

    if n_fft < n_persegs:
        n_fft = n_persegs
        logging.warning("nfft < n_persegs. set to n_persegs")

    f_samp = 1e9 / np.median(np.diff(inp.time.data)).astype(np.float64)

    freqs, p_xx = signal.welch(
        inp.data,
        nfft=n_fft,
        fs=f_samp,
        window=window,
        noverlap=n_overlap,
        detrend=d_flag,
        nperseg=n_persegs,
        scaling=scaling,
        return_onesided=True,
        axis=-1,
    )

    out = xr.DataArray(p_xx, coords=[freqs], dims=["f"])

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import xarray as xr

# Local imports
from .cotrans import cotrans

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def gse2gsm(inp, flag: str = "gse>gsm"):
    r"""Converts GSE to GSM.

    Parameters
    ----------
    inp : xarray.DataArray or ndarray
        Time series of the input in GSE (GSM) coordinates.
        If ndarray first column is time in unix format.
    flag : {"gse>gsm", "gsm>gse"}, Optional
        Flag for conversion direction. Default is "gse>gsm"

    Returns
    -------
    out : xarray.DataArray or ndarray
        Time series of the input in GSM (GSE) coordinates.
        If ndarray first column is time in unix format.

    See also
    --------
    pyrfu.pyrf.geocentric_coordinate_transformation

    """

    assert isinstance(inp, xr.DataArray), "inp must be a xarray.DataArray"
    assert inp.ndim == 2 and inp.shape[1] == 3, "inp must be a vector"

    message = "flag must be a string gse>gsm or gsm>gse"
    assert isinstance(flag, str) and flag.lower() in ["gse>gsm", "gsm>gse"], message

    out = cotrans(inp, flag)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def eis_omni(eis_allt, method: str = "mean"):
    r"""Calculates the omni-directional flux for all 6 telescopes.

    Parameters
    ----------
    eis_allt : xarray.Dataset
        Dataset of the fluxes of all 6 telescopes.

    Returns
    -------
    flux_omni : xarray.DataArray
        Omni-directional flux for all 6 telescopes

    See Also
    --------
    pyrfu.mms.get_eis_allt

    Examples
    --------
    >>> from pyrfu import mms

    Define spacecraft index and time interval

    >>> tint = ["2017-07-23T16:10:00", "2017-07-23T18:10:00"]
    >>> ic = 2

    Get EIS ExTOF all 6 telescopes fluxes

    >>> extof_allt = mms.get_eis_allt("flux_extof_proton_srvy_l2", tint, ic)

    Compute the omni-directional flux for all 6 telescopes

    >>> extof_omni = mms.eis_omni(extof_allt)

    """

    assert method.lower() in ["mean", "sum"]

    scopes = list(filter(lambda x: x[0] == "t", eis_allt))

    flux_omni = np.zeros_like(eis_allt[scopes[0]].data)

    for scope in scopes:
        flux_omni += eis_allt[scope].data.copy()

        # Why??
        # try:
        #     flux_omni += eis_allt[scope].data.copy()
        # except TypeError:
        #     flux_omni = eis_allt[scope].data.copy()

    if method.lower() == "mean":
        flux_omni /= len(scopes)

    # Get dimensions, coordinates and attributes based on first telescope
    dims = eis_allt[scopes[0]].dims
    coords = [eis_allt[scopes[0]][k] for k in dims]
    attrs = eis_allt[scopes[0]].attrs

    flux_omni = xr.DataArray(flux_omni, coords=coords, dims=dims, attrs=attrs)

    return flux_omni



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from cdflib import cdfread

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def get_variable(file_path, cdf_name):
    r"""Reads field named cdf_name in file and convert to DataArray.

    Parameters
    ----------
    file_path : str
        Path of the cdf file.
    cdf_name : str
        Name of the target variable in the cdf file.

    Returns
    -------
    out : xarray.DataArray
        Target variable.

    """

    with cdfread.CDF(file_path) as file:
        var_data = file.varget(cdf_name)
        var_atts = file.varattsget(cdf_name)

    out = xr.DataArray(
        var_data,
        coords=[np.arange(len(var_data))],
        dims=["x"],
        attrs=var_atts,
    )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

from ..pyrf.ts_skymap import ts_skymap

# Local imports
from .dpf2psd import dpf2psd

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def eis_skymap(inp_ang_ang, to_psd: bool = True):
    r"""Construct skymap distribution from angle-angle distribution.

    Parameters
    ----------
    inp_ang_ang : xarray.DataArray
        EIS angle-angle distribution.
    to_psd : bool, Optional
        Flag to convert differential particle flux to phase space density.

    Returns
    -------
    out : xarray.Dataset
        EIS skymap distribution.

    See Also
    --------
    pyrfu.mms.eis_ang_ang

    """

    time_data = inp_ang_ang.time.data
    energy = np.tile(inp_ang_ang.energy.data, (len(time_data), 1))
    phi = np.tile(inp_ang_ang.phi.data, (len(time_data), 1))
    theta = inp_ang_ang.theta.data

    a = inp_ang_ang.attrs
    a = list(
        filter(
            lambda k: k not in ["delta_energy_plus", "delta_energy_minus"],
            a,
        ),
    )
    attrs = {k: inp_ang_ang.attrs[k] for k in a}
    coords_attrs = {k: inp_ang_ang[k].attrs for k in ["time", "energy", "phi", "theta"]}

    glob_attrs = inp_ang_ang.attrs["GLOBAL"]
    glob_attrs = {
        "delta_energy_plus": inp_ang_ang.attrs["delta_energy_plus"],
        "delta_energy_minus": inp_ang_ang.attrs["delta_energy_minus"],
        "species": inp_ang_ang.attrs["species"],
        **glob_attrs,
    }

    energy0, energy1 = [energy[i, :] for i in range(2)]
    e_step_table = np.zeros(len(time_data))

    out = ts_skymap(
        time_data,
        inp_ang_ang.data,
        energy,
        phi,
        theta,
        energy0=energy0,
        energy1=energy1,
        esteptable=e_step_table,
        attrs=attrs,
        coords_attrs=coords_attrs,
        glob_attrs=glob_attrs,
    )
    # out.attrs["species"] = "ions"
    # out.attrs["UNITS"] = "1/(cm^2 s sr keV)"

    out.energy.data *= 1e3

    if to_psd:
        out = dpf2psd(out)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import logging

# 3rd party imports
import numba
import numpy as np
import xarray as xr
from scipy import constants

# Local imports
from ..pyrf.resample import resample
from ..pyrf.ts_scalar import ts_scalar
from ..pyrf.ts_tensor_xyz import ts_tensor_xyz
from ..pyrf.ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


@numba.jit(cache=True, fastmath=True, nogil=True, parallel=True, nopython=True)
def _moms(
    energy,
    delta_v,
    q_e,
    sc_pot,
    p_mass,
    flag_inner_electron,
    w_inner_electron,
    phi,
    theta,
    int_energies,
    vdf,
    delta_ang,
):
    n_psd = np.zeros(vdf.shape[0])
    v_psd = np.zeros((vdf.shape[0], 3))
    p_psd = np.zeros((vdf.shape[0], 3, 3))
    h_psd = np.zeros((vdf.shape[0], 3))

    for i_t in numba.prange(vdf.shape[0]):
        energy_correct = energy[i_t, :] - sc_pot[i_t]

        velocity = np.sqrt(2 * q_e * energy_correct / p_mass)
        velocity[energy_correct < flag_inner_electron * w_inner_electron] = 0

        phi_i = np.deg2rad(phi[i_t, :, :])
        theta_i = np.deg2rad(theta[i_t, :, :])

        psd2n_mat = np.ones(theta_i.shape) * np.sin(theta_i)

        # Particle flux and heat flux vector
        psd2v_x_mat = -np.cos(phi_i) * np.sin(theta_i) ** 2
        psd2v_y_mat = -np.sin(phi_i) * np.sin(theta_i) ** 2
        psd2v_z_mat = -np.ones(theta_i.shape) * np.sin(theta_i) * np.cos(theta_i)

        psd2p_xx_mat = np.cos(phi_i) ** 2.0 * np.sin(theta_i) ** 3
        psd2p_yy_mat = np.sin(phi_i) ** 2.0 * np.sin(theta_i) ** 3
        psd2p_zz_mat = np.ones(theta_i.shape) * np.sin(theta_i) * np.cos(theta_i) ** 2
        psd2p_xy_mat = np.cos(phi_i) * np.sin(phi_i) * np.sin(theta_i) ** 3
        psd2p_xz_mat = np.cos(phi_i) * np.sin(phi_i) ** 2 * np.cos(theta_i)
        psd2p_yz_mat = np.sin(phi_i) * np.sin(phi_i) ** 2 * np.cos(theta_i)

        for i_e in int_energies:
            tmp = vdf[i_t, i_e, :, :]
            # n_psd_tmp1 = tmp .* psd2_n_mat * v(ii)^2 * delta_v(ii) * delta_ang;
            # n_psd_e32_phi_theta(nt, ii, :, :) = n_psd_tmp1;
            # n_psd_e32(nt, ii) = n_psd_tmp

            # number density
            n_psd_tmp = np.nansum(tmp * psd2n_mat * delta_ang[i_t])
            n_psd_tmp *= delta_v[i_t, i_e] * velocity[i_e] ** 2
            n_psd[i_t] += n_psd_tmp

            # Bulk velocity
            v_temp_x = np.nansum(tmp * psd2v_x_mat * delta_ang[i_t])
            v_temp_x *= delta_v[i_t, i_e] * velocity[i_e] ** 3

            v_temp_y = np.nansum(tmp * psd2v_y_mat * delta_ang[i_t])
            v_temp_y *= delta_v[i_t, i_e] * velocity[i_e] ** 3

            v_temp_z = np.nansum(tmp * psd2v_z_mat * delta_ang[i_t])
            v_temp_z *= delta_v[i_t, i_e] * velocity[i_e] ** 3

            v_psd[i_t, 0] += v_temp_x
            v_psd[i_t, 1] += v_temp_y
            v_psd[i_t, 2] += v_temp_z

            # Pressure tensor
            p_temp_xx = np.nansum(tmp * psd2p_xx_mat * delta_ang[i_t])
            p_temp_xx *= delta_v[i_t, i_e] * velocity[i_e] ** 4

            p_temp_xy = np.nansum(tmp * psd2p_xy_mat * delta_ang[i_t])
            p_temp_xy *= delta_v[i_t, i_e] * velocity[i_e] ** 4

            p_temp_xz = np.nansum(tmp * psd2p_xz_mat * delta_ang[i_t])
            p_temp_xz *= delta_v[i_t, i_e] * velocity[i_e] ** 4

            p_temp_yy = np.nansum(tmp * psd2p_yy_mat * delta_ang[i_t])
            p_temp_yy *= delta_v[i_t, i_e] * velocity[i_e] ** 4

            p_temp_yz = np.nansum(tmp * psd2p_yz_mat * delta_ang[i_t])
            p_temp_yz *= delta_v[i_t, i_e] * velocity[i_e] ** 4

            p_temp_zz = np.nansum(tmp * psd2p_zz_mat * delta_ang[i_t])
            p_temp_zz *= delta_v[i_t, i_e] * velocity[i_e] ** 4

            p_psd[i_t, 0, 0] += p_temp_xx
            p_psd[i_t, 0, 1] += p_temp_xy
            p_psd[i_t, 0, 2] += p_temp_xz
            p_psd[i_t, 1, 1] += p_temp_yy
            p_psd[i_t, 1, 2] += p_temp_yz
            p_psd[i_t, 2, 2] += p_temp_zz

            h_psd[i_t, 0] = v_temp_x * velocity[i_e] ** 2
            h_psd[i_t, 1] = v_temp_y * velocity[i_e] ** 2
            h_psd[i_t, 2] = v_temp_z * velocity[i_e] ** 2

    return n_psd, v_psd, p_psd, h_psd


def psd_moments(vdf, sc_pot, **kwargs):
    r"""Computes moments from the FPI particle phase-space densities.

    Parameters
    ----------
    vdf : xarray.Dataset
        3D skymap velocity distribution.
    sc_pot : xarray.DataArray
        Time series of the spacecraft potential.

    Returns
    -------
    n_psd : xarray.DataArray
        Time series of the number density (1rst moment).
    v_psd : xarray.DataArray
        Time series of the bulk velocity (2nd moment).
    p_psd : xarray.DataArray
        Time series of the pressure tensor (3rd moment).
    p2_psd : xarray.DataArray
        Time series of the pressure tensor.
    t_psd : xarray.DataArray
        Time series of the temperature tensor.
    h_psd : xarray.DataArray
        to fill.

    Other Parameters
    ----------------
    energy_range : array_like
        Set energy range in eV to integrate over [E_min E_max]. Energy range
        is applied to energy0 and the same elements are used for energy1 to
        ensure that the same number of points are integrated over.
    no_sc_pot : bool
        Set to 1 to set spacecraft potential to zero. Calculates moments
        without correcting for spacecraft potential.
    en_channels : array_like
        Set energy channels to integrate over [min max]; min and max between
        must be between 1 and 32.
    partial_moments : numpy.ndarray or xarray.DataArray
        Use a binary array to select which psd points are used in the moments
        calculation. `partial_moments` must be a binary array (1s and 0s,
        1s correspond to points used). Array (or data of Dataarray) must be the same
        size as vdf.data.
    inner_electron : {"on", "off"}
        inner_electrontron potential for electron moments.

    Examples
    --------
    >>> from pyrfu import mms

    Define time interval

    >>> tint_brst = ["2015-10-30T05:15:20.000", "2015-10-30T05:16:20.000"]

    Load magnetic field and spacecraft potential

    >>> scpot = mms.get_data("V_edp_brst_l2", tint_brst, 1)

    Load electron velocity distribution function

    >>> vdf_e = mms.get_data("pde_fpi_brst_l2", tint_brst, 1)

    Compute moments

    >>> options = dict(energy_range=[1, 1000])
    >>> moments_e = mms.psd_moments(vdf_e, scpot, **options)

    """

    # [eV] sc_pot + w_inner_electron for electron moments calculation
    w_inner_electron = 3.5

    # Check if data is fast or burst resolution
    if "brst" in vdf.data.attrs["FIELDNAM"].lower():
        is_brst_data = True
        logging.info("Burst resolution data is used")
    elif "fast" in vdf.data.attrs["FIELDNAM"].lower():
        is_brst_data = False
        logging.info("Fast resolution data is used")
    else:
        raise TypeError("Could not identify if data is fast or burst.")

    theta = vdf.theta.data
    particle_type = vdf.attrs["species"]
    assert particle_type[0].lower() in ["e", "i"], "invalid particle type"

    vdf_data = vdf.data.data * 1e12  # In SI units

    step_table = vdf.attrs["esteptable"]
    energy = vdf.energy.data
    energy0 = vdf.attrs["energy0"]
    energy1 = vdf.attrs["energy1"]
    e_tmp = energy1 - energy0

    flag_same_e = all(e_tmp) == 0

    # resample sc_pot to same resolution as particle distributions
    sc_pot = resample(sc_pot, vdf.time).data

    if "energy_range" in kwargs:
        if (
            isinstance(kwargs["energy_range"], (list, np.ndarray))
            and len(kwargs["energy_range"]) == 2
        ):
            # if not is_brst_data:
            #    energy0 = energy

            # e_min_max = kwargs["energy_range"]
            # start_e = bisect.bisect_left(energy0, e_min_max[0])
            # stop_e = bisect.bisect_left(energy0, e_min_max[1])

            logging.info("Using partial energy range")

    no_sc_pot = kwargs.get("no_sc_pot", False)
    if no_sc_pot:
        sc_pot = np.zeros(sc_pot.shape)
        logging.info("Setting spacecraft potential to zero")

    int_energies = np.arange(
        kwargs.get("en_channels", [0, 32])[0],
        kwargs.get("en_channels", [0, 32])[1],
    )

    if "partial_moments" in kwargs:
        partial_moments = kwargs["partial_moments"]
        if isinstance(partial_moments, xr.DataArray):
            partial_moments = partial_moments.data

        # Check size of partial_moments
        if partial_moments.shape == vdf_data.shape:
            sum_ones = np.sum(
                np.sum(
                    np.sum(np.sum(partial_moments, axis=-1), axis=-1),
                    axis=-1,
                ),
                axis=-1,
            )
            sum_zeros = np.sum(
                np.sum(
                    np.sum(np.sum(-partial_moments + 1, axis=-1), axis=-1),
                    axis=-1,
                ),
                axis=-1,
            )

            if (sum_ones + sum_zeros) == vdf_data.size:
                logging.info(
                    "partial_moments is correct. Partial moments will be calculated"
                )
                vdf_data = vdf_data * partial_moments
            else:
                logging.info(
                    "All values are not ones and zeros in partial_moments. "
                    "Full moments will be calculated"
                )
        else:
            logging.info(
                "Size of partial_moments is wrong. Full moments will be calculated"
            )

    tmp_ = kwargs.get("inner_electron", "")
    flag_inner_electron = tmp_ == "on" and particle_type[0] == "e"

    # Define constants
    q_e = constants.elementary_charge
    k_b = constants.Boltzmann

    if particle_type[0] == "e":
        p_mass = constants.electron_mass
        logging.info("Particles are electrons")
    else:
        p_mass = constants.proton_mass
        sc_pot *= -1.0
        logging.info("Particles are ions")

    # angle between theta and phi points is 360/32 = 11.25 degrees
    phi = vdf.phi.data

    if "delta_phi_minus" in vdf.attrs and "delta_phi_plus" in vdf.attrs:
        delta_phi_minus = vdf.attrs["delta_phi_minus"]
        delta_phi_plus = vdf.attrs["delta_phi_plus"]
        delta_phi = delta_phi_plus + delta_phi_minus
        delta_phi = np.tile(delta_phi[:, :, np.newaxis], (1, 1, vdf_data.shape[3]))
    else:
        delta_phi = np.deg2rad(np.median(np.diff(phi[0, :])))
        delta_phi = delta_phi * np.ones(
            (vdf_data.shape[0], vdf_data.shape[2], vdf_data.shape[3])
        )

    if "delta_theta_minus" in vdf.attrs and "delta_theta_plus" in vdf.attrs:
        delta_theta_minus = vdf.attrs["delta_theta_minus"]
        delta_theta_plus = vdf.attrs["delta_theta_plus"]
        delta_theta = delta_theta_plus + delta_theta_minus
        delta_theta = np.tile(
            delta_theta[np.newaxis, np.newaxis, :],
            (vdf_data.shape[0], vdf_data.shape[2], 1),
        )
    else:
        delta_theta = np.deg2rad(np.median(np.diff(theta)))
        delta_theta = delta_theta * np.ones(
            (vdf_data.shape[0], vdf_data.shape[2], vdf_data.shape[3])
        )

    delta_ang = delta_phi * delta_theta

    phi_mat = np.tile(phi[:, :, np.newaxis], (1, 1, vdf_data.shape[3]))
    theta_mat = np.tile(
        theta[np.newaxis, np.newaxis, :], (vdf_data.shape[0], vdf_data.shape[2], 1)
    )

    energy_minus = vdf.attrs["delta_energy_plus"]
    energy_plus = vdf.attrs["delta_energy_plus"]

    energy_correct = energy - sc_pot[:, np.newaxis]
    velocity = np.sqrt(2 * q_e * energy_correct / p_mass)
    velocity[energy_correct < flag_inner_electron * w_inner_electron] = 0

    # Calculate speed widths associated with each energy channel.
    if is_brst_data:  # Burst mode energy/speed widths
        if flag_same_e:
            energy_upper = energy + energy_plus
            energy_lower = energy - energy_minus
            v_upper = np.sqrt(2 * q_e * energy_upper / p_mass)
            v_lower = np.sqrt(2 * q_e * energy_lower / p_mass)
            delta_v = v_upper - v_lower
            delta_v = np.tile(delta_v, (vdf_data.shape[0], 1))
        else:
            energy_all = np.hstack([energy0, energy1])
            energy_all = np.log10(np.sort(energy_all))

            if np.abs(energy_all[1] - energy_all[0]) > 1e-4:
                temp0 = 2 * energy_all[0] - energy_all[1]
            else:
                temp0 = 2 * energy_all[1] - energy_all[2]

            if np.abs(energy_all[63] - energy_all[62]) > 1e-4:
                temp65 = 2 * energy_all[63] - energy_all[62]
            else:
                temp65 = 2 * energy_all[63] - energy_all[61]

            energy_all = np.hstack([temp0, energy_all, temp65])
            diff_en_all = np.diff(energy_all)
            energy0upper = 10 ** (np.log10(energy0) + diff_en_all[1:64:2] / 2)
            energy0lower = 10 ** (np.log10(energy0) - diff_en_all[0:63:2] / 2)
            energy1upper = 10 ** (np.log10(energy1) + diff_en_all[2:65:2] / 2)
            energy1lower = 10 ** (np.log10(energy1) - diff_en_all[1:64:2] / 2)

            v0upper = np.sqrt(2 * q_e * energy0upper / p_mass)
            v0lower = np.sqrt(2 * q_e * energy0lower / p_mass)
            v1upper = np.sqrt(2 * q_e * energy1upper / p_mass)
            v1lower = np.sqrt(2 * q_e * energy1lower / p_mass)
            delta_v0 = (v0upper - v0lower) * 2.0
            delta_v1 = (v1upper - v1lower) * 2.0

            delta_v = np.tile(delta_v0, (vdf_data.shape[0], 1))
            delta_v[step_table == 1, :] = delta_v1

    else:  # Fast mode energy/speed widths
        energy_all = np.log10(energy[0, :])
        temp0 = 2 * energy_all[0] - energy_all[1]
        temp33 = 2 * energy_all[31] - energy_all[30]
        energy_all = np.hstack([temp0, energy_all, temp33])
        diff_en_all = np.diff(energy_all)
        energy_upper = 10 ** (np.log10(energy[0, :]) + diff_en_all[1:33] / 4)
        energy_lower = 10 ** (np.log10(energy[0, :]) - diff_en_all[0:32] / 4)
        v_upper = np.sqrt(2 * q_e * energy_upper / p_mass)
        v_lower = np.sqrt(2 * q_e * energy_lower / p_mass)
        delta_v = (v_upper - v_lower) * 2.0
        delta_v[0] = delta_v[0] * 2.7
        delta_v = np.tile(delta_v, (vdf_data.shape[0], 1))

    n_psd, v_psd, p_psd, h_psd = _moms(
        energy,
        delta_v,
        q_e,
        sc_pot,
        p_mass,
        flag_inner_electron,
        w_inner_electron,
        phi_mat,
        theta_mat,
        int_energies,
        vdf_data,
        delta_ang,
    )

    # Compute moments in SI units
    p_psd *= p_mass
    v_psd /= n_psd[:, np.newaxis]
    p2_psd = np.zeros_like(p_psd)
    p2_psd[:, 0, 0] = p_psd[:, 0, 0]
    p2_psd[:, 0, 1] = p_psd[:, 0, 1]
    p2_psd[:, 0, 2] = p_psd[:, 0, 2]
    p2_psd[:, 1, 1] = p_psd[:, 1, 1]
    p2_psd[:, 1, 2] = p_psd[:, 1, 2]
    p2_psd[:, 2, 2] = p_psd[:, 2, 2]
    p2_psd[:, 1, 0] = p2_psd[:, 0, 1]
    p2_psd[:, 2, 0] = p2_psd[:, 0, 2]
    p2_psd[:, 2, 1] = p2_psd[:, 1, 2]

    p_psd[:, 0, 0] -= p_mass * n_psd * v_psd[:, 0] * v_psd[:, 0]
    p_psd[:, 0, 1] -= p_mass * n_psd * v_psd[:, 0] * v_psd[:, 1]
    p_psd[:, 0, 2] -= p_mass * n_psd * v_psd[:, 0] * v_psd[:, 2]
    p_psd[:, 1, 1] -= p_mass * n_psd * v_psd[:, 1] * v_psd[:, 1]
    p_psd[:, 1, 2] -= p_mass * n_psd * v_psd[:, 1] * v_psd[:, 2]
    p_psd[:, 2, 2] -= p_mass * n_psd * v_psd[:, 2] * v_psd[:, 2]
    p_psd[:, 1, 0] = p_psd[:, 0, 1]
    p_psd[:, 2, 0] = p_psd[:, 0, 2]
    p_psd[:, 2, 1] = p_psd[:, 1, 2]

    p_trace = np.trace(p_psd, axis1=1, axis2=2)
    t_psd = np.zeros(p_psd.shape)
    t_psd[...] = p_psd[...] / (k_b * n_psd[:, np.newaxis, np.newaxis])
    t_psd[:, 1, 0] = t_psd[:, 1, 0]
    t_psd[:, 2, 0] = t_psd[:, 2, 0]
    t_psd[:, 2, 1] = t_psd[:, 2, 1]

    v_abs2 = np.linalg.norm(v_psd, axis=1) ** 2
    h_psd *= p_mass / 2
    h_psd[:, 0] -= v_psd[:, 0] * p_psd[:, 0, 0]
    h_psd[:, 0] -= v_psd[:, 1] * p_psd[:, 0, 1]
    h_psd[:, 0] -= v_psd[:, 2] * p_psd[:, 0, 2]
    h_psd[:, 0] -= 0.5 * v_psd[:, 0] * (p_trace + p_mass * n_psd * v_abs2)
    h_psd[:, 1] -= v_psd[:, 0] * p_psd[:, 1, 0]
    h_psd[:, 1] -= v_psd[:, 1] * p_psd[:, 1, 1]
    h_psd[:, 1] -= v_psd[:, 2] * p_psd[:, 1, 2]
    h_psd[:, 1] -= 0.5 * v_psd[:, 1] * (p_trace + p_mass * n_psd * v_abs2)
    h_psd[:, 2] -= v_psd[:, 0] * p_psd[:, 2, 0]
    h_psd[:, 2] -= v_psd[:, 1] * p_psd[:, 2, 1]
    h_psd[:, 2] -= v_psd[:, 2] * p_psd[:, 2, 2]
    h_psd[:, 2] -= 0.5 * v_psd[:, 2] * (p_trace + p_mass * n_psd * v_abs2)

    # Convert to typical units (/cc, km/s, nP, eV, and ergs/s/cm^2).
    n_psd /= 1e6
    v_psd /= 1e3
    p_psd *= 1e9
    p2_psd *= 1e9
    t_psd *= k_b / q_e
    h_psd *= 1e3

    # Construct TSeries
    n_psd = ts_scalar(vdf.time.data, n_psd)
    v_psd = ts_vec_xyz(vdf.time.data, v_psd)
    p_psd = ts_tensor_xyz(vdf.time.data, p_psd)
    p2_psd = ts_tensor_xyz(vdf.time.data, p2_psd)
    t_psd = ts_tensor_xyz(vdf.time.data, t_psd)
    h_psd = ts_vec_xyz(vdf.time.data, h_psd)

    return n_psd, v_psd, p_psd, p2_psd, t_psd, h_psd



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

from .dot import dot
from .norm import norm

# Local imports
from .resample import resample

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def pvi_4sc(b_mms):
    r"""Compute the Partial Variance of Increments (PVI) using the
    definition in [1]_ as

    .. math::

            PVI_{ij}(t) = \sqrt{\frac{|\Delta B_{ij}(t)|^2}
            {\langle|\Delta B_{ij}|^2\rangle}}

    where :math:`\Delta B_{ij}(t) = B_i(t) - B_i(t)` is the magnetic field
    increments, the average :math:`\langle . \rangle` is taken over the
    whole interval, and :math:`i`, :math:`j` = 1,2,3,4 is the MMS spacecraft
    number.

    In addition, computes, the rotation of the magnetic field between
    two spacecraft, i.e., magnetic field shear angle, as :

    .. math::

        \alpha_{ij}(t) = cos^{-1} \frac{B_i(t) . B_j(t)}{|B_i(t)| |B_j(t)|}


    Parameters
    ----------
    b_mms : list of xarray.DataArray
        List of the time series of the background magnetic field for
        the 4 spacecraft.

    Returns
    -------
    pvi_ij : xarray.DataArray
        Time series of the Partial Variance of Increments for the 6
        pairs of spacecraft.
    alpha_ij : xarray.DataArray
        Time series of the magnetic field shear angle for the 6 pairs
        of spacecraft.

    References
    ----------
    .. [1] Chasapis, A., Retinó, A., Sahraoui, F., Vaivads, A.,
            Khotyaintsev, Yu. V., Sundkvist, D., et al. (2015)
            Thin current sheets and associated electron heating in
            turbulent space plasma. Astrophys. J. Lett. 804:L1.
            doi: https://doi.org/10.1088/2041-8205/804/1/L1


    """

    b_mms = [resample(b_xyz, b_mms[0]) for b_xyz in b_mms]

    i_indices = [0, 0, 0, 1, 1, 2]
    j_indices = [1, 2, 3, 2, 3, 3]

    # Compute normalized partial variance of increments
    def pvi(b_i, b_j):
        return np.sqrt(norm(b_i - b_j) ** 2 / np.mean(norm(b_i - b_j) ** 2))

    pvi_ij = [pvi(b_mms[i], b_mms[j]) for i, j in zip(i_indices, j_indices)]

    # Compute magnetic field shear angle
    def shear_angle(b_i, b_j):
        return np.arccos(dot(b_i, b_j) / (norm(b_i) * norm(b_j)))

    theta_ij = []
    for i, j in zip(i_indices, j_indices):
        theta_ij.append(shear_angle(b_mms[i], b_mms[j]))

    return pvi_ij, theta_ij



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .disp_surf_calc import disp_surf_calc

# @Louis Richard
from .one_fluid_dispersion import one_fluid_dispersion

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

__all__ = ["disp_surf_calc", "one_fluid_dispersion"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json

# Built-in imports
import os

# Thrid party imports
import numpy as np
import xarray as xr
from scipy import constants, interpolate, optimize
from scipy.spatial.transform import Rotation as R

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

__all__ = ["shock_normal"]


def shock_normal(spec, leq90: bool = True):
    r"""Calculates shock normals with different methods. Normal vectors are calculated
    by methods described in [1]_ and references therein.

    The data can be averaged values or values from the time series in matrix format.
    If series is from time series all parameters are calculated from a random
    upstream and a random downstream point. This can help set errorbars on shock
    angle etc. The time series input must have the same size (up- and downstream can
    be different), so generally the user needs to resample the data first.

    Parameters
    ----------
    spec : dict
        Hash table with:
            * b_u : Upstream magnetic field (nT).
            * b_d : Downstream magnetic field.
            * v_u : Upstream plasma bulk velocity (km/s).
            * v_d : Downstream plasma bulk velocity.
            * n_u : Upstream number density (cm^-3).
            * n_d : Downstream number density.
            * r_xyz : Spacecraft position in time series format of 1x3 vector. Optional.
            * d2u : Down-to-up, is 1 or -1. Optional.
            * dt_f : Time duration of shock foot (s). Optional.
            * f_cp : Reflected ion gyrofrequency (Hz). Optional.
            * n : Number of Monte Carlo particles. Optional, default is 100.

    leq90 : bool, Optional
        Force angles to be less than 90 (default). For leq90 = 0, angles can be between
        0 and 180 deg. For time series input and quasi-perp shocks,leq90 = 0 is
        recommended.

    Returns
    -------
    out : dict
        Hash table with:
            * n : Hash table containing normal vectors (n always points toward the
            upstream region).
            From data:
                * mc : Magnetic coplanarity (10.14)
                * vc : Velocity coplanarity (10.18)
                * mx_1 : Mixed method 1 (10.15), [2]_
                * mx_2 : Mixed method 2 (10.16), [2]_
                * mx_3 : Mixed method 3 (10.17), [2]_
            From models (only if r_xyz is included in spec):
                * farris : [3]_
                * slho : [4]_
                * per : [5]_, (z = 0)
                * fa4o : [6]_
                * fan4o : [6]_
                * foun : [7]_

            * theta_bn : Angle between normal vector and b_u, same fields as n.
            * theta_vn : Angle between normal vector and v_u, same fields as n.
            * v_sh : Hash table containing shock velocities:
                * gt : Using shock foot thickness (10.32). [8]_
                * mf : Mass flux conservation (10.29).
                * sb : Using jump conditions (10.33). [9]_
                * mo : Using shock foot thickness
            * info : Hash table containing some more info:
                * msh : Magnetic shear angle.
                * vsh : Velocity shear angle.
                * cmat : Constraints matrix with normalized errors.
                * sig : Scaling factor to fit shock models to sc position. Calculated
                from (10.9-10.13) in [1]_


    References
    ----------
    .. [1]  Schwartz, S. J. (1998), Shock and Discontinuity Normals, Mach Numbers, and
            Related Parameters, ISSI Scientific Reports Series, vol. 1, pp. 249–270.
    .. [2]  Abraham-Shrauner, B. (1972), “Determination of magnetohydrodynamic shock
            normals”, Journal of Geophysical Research, vol. 77, no. 4, p. 736.
            doi:10.1029/JA077i004p00736.
    .. [3]  Farris, M. H., Petrinec, S. M., and Russell, C. T. (1991), The thickness of
            the magnetosheath: Constraints on the polytropic index”, Geophysical
            Research Letters, vol. 18, no. 10, pp. 1821–1824. doi:10.1029/91GL02090.
    .. [4]  Slavin, J. A. and Holzer, R. E. (1981), Solar wind flow about the
            terrestrial planets, 1. Modeling bow shock position and shape, Journal of
            Geophysical Research, vol. 86, no. A13, pp. 11401–11418.
            doi:10.1029/JA086iA13p11401.
    .. [5]  Peredo, M., Slavin, J. A., Mazur, E., and Curtis, S. A. (1995),
            Three-dimensional position and shape of the bow shock and their variation
            with Alfvénic, sonic, and magnetosonic Mach numbers and interplanetary
            magnetic field orientation, Journal of Geophysical Research, vol. 100,
            no. A5, pp. 7907–7916. doi:10.1029/94JA02545.
    .. [6]  Fairfield, D. H. (1971), Average and unusual locations of the Earth's
            magnetopause and bow shock, Journal of Geophysical Research, vol. 76,
            no. 28, p. 6700, 1971. doi:10.1029/JA076i028p06700.
    .. [7]  Formisano, V. (1979), Orientation and Shape of the Earth's Bow Shock in
            Three Dimensions, Planetary and Space Science, vol. 27, no. 9,
            pp. 1151–1161. doi:10.1016/0032-0633(79)90135-1.
    .. [8]  Gosling, J. T. and Thomsen, M. F. (1985), Specularly reflected ions, shock
            foot thicknesses, and shock velocity determination in space, Journal of
            Geophysical Research, vol. 90, no. A10, pp. 9893–9896.
            doi:10.1029/JA090iA10p09893.
    .. [9]  Smith, E. J. and Burton, M. E. (1988), Shock analysis: Three useful new
            relations, Journal of Geophysical Research, vol. 93, no. A4, pp. 2730–2734.
            doi:10.1029/JA093iA04p02730.



    """

    # Check input
    assert isinstance(spec, dict), "spec must be a dictionary"

    if spec["b_u"].ndim > 1 or spec["b_d"].ndim > 1:
        n_bu = len(spec["b_u"])
        n_bd = len(spec["b_d"])

        # randomize points upstream and downstream
        n = int(np.floor(spec.get("n", 10.0)))
        idt_u, idt_d = [np.random.rand(n) for _ in range(2)]

        tmp_spec = {}
        for i in range(n):
            f_bu = interpolate.interp1d(np.linspace(0, 1, n_bu), spec["b_u"], axis=0)
            tmp_spec["b_u"] = f_bu(idt_u[i])
            f_vu = interpolate.interp1d(np.linspace(0, 1, n_bu), spec["v_u"], axis=0)
            tmp_spec["v_u"] = f_vu(idt_u[i])
            f_nu = interpolate.interp1d(np.linspace(0, 1, n_bu), spec["n_u"], axis=0)
            tmp_spec["n_u"] = f_nu(idt_u[i])

            f_bd = interpolate.interp1d(np.linspace(0, 1, n_bd), spec["b_d"], axis=0)
            tmp_spec["b_d"] = f_bd(idt_d[i])
            f_vd = interpolate.interp1d(np.linspace(0, 1, n_bd), spec["v_d"], axis=0)
            tmp_spec["v_d"] = f_vd(idt_d[i])
            f_nd = interpolate.interp1d(np.linspace(0, 1, n_bd), spec["n_d"], axis=0)
            tmp_spec["n_d"] = f_nd(idt_d[i])

    # normal vector, according to different models
    normal = {}

    b_u, b_d = np.array(spec["b_u"]), np.array(spec["b_d"])
    v_u, v_d = np.array(spec["v_u"]), np.array(spec["v_d"])

    delta_b = b_d - b_u
    delta_v = v_d - v_u
    spec["delta_b"] = delta_b
    spec["delta_v"] = delta_v

    # magenetic coplanarity
    mc = np.cross(np.cross(b_d, b_u), delta_b)
    mc /= np.linalg.norm(mc, keepdims=True)
    normal["mc"] = mc

    # velocity coplanarity
    vc = delta_v / np.linalg.norm(delta_v, keepdims=True)
    normal["vc"] = vc

    # mixed methods
    mx_1 = np.cross(np.cross(b_u, delta_v), delta_b)
    mx_1 /= np.linalg.norm(mx_1)
    normal["mx_1"] = mx_1
    mx_2 = np.cross(np.cross(b_d, delta_v), delta_b)
    mx_2 /= np.linalg.norm(mx_2)
    normal["mx_2"] = mx_2
    mx_3 = np.cross(np.cross(delta_b, delta_v), delta_b)
    mx_3 /= np.linalg.norm(mx_3)
    normal["mx_3"] = mx_3

    # Load shock
    # pkg_path = os.path.join(os.getcwd(), "sandbox", "sbox")
    pkg_path = os.path.dirname(os.path.abspath(__file__))
    with open(
        os.path.join(pkg_path, "shock_models_parameters.json"), "r", encoding="utf-8"
    ) as fs:
        shock_models_params = json.load(fs)

    if "r_xyz" in spec:
        # info
        info = {k: {} for k in shock_models_params["farris"]}
        sig = {}

        # overwrite alpha to azimuthal angle of the bulk velocity for
        # Slavin-Holzer model
        alpha_sh = -np.rad2deg(np.arctan(spec["v_u"][1] / spec["v_u"][0]))
        shock_models_params["slavin_holzer"]["alpha"] = alpha_sh

        for m in shock_models_params:
            for k in info:
                info[k][m] = shock_models_params[m][k]

            normal[m], sig[m] = _shock_model(spec, *shock_models_params[m].values())

        info["sig"] = sig
    else:
        info = {}

    # make sure all normal vectors are pointing upstream based on delta_sv,
    # should work for IP shocks also
    for n_ in normal.values():
        if np.sum(delta_v * n_) < 0:
            n_ *= -1

    # Shock normal to magnetic field and velocity angles
    theta_bn = _shock_angle(spec, normal, "b", leq90)
    theta_vn = _shock_angle(spec, normal, "v", leq90)

    # Magnetic and velocity shear angles
    info["msh"] = _shear_angle(b_u, b_d)
    info["vsh"] = _shear_angle(v_u, v_d)

    # Constraint matrix
    info["cmat"] = _constraint_matrix(spec, normal)

    # Shock speed
    v_sh = {}
    for m in ["gt", "mf", "sb", "mo"]:
        v_sh[m] = _shock_speed(spec, normal, theta_bn, m)

    out = {
        "info": info,
        "n": normal,
        "theta_bn": theta_bn,
        "theta_vn": theta_vn,
        "v_sh": v_sh,
    }

    return out


def _shear_angle(au, ad):
    theta = np.arccos(np.sum(au * ad) / (np.linalg.norm(au) * np.linalg.norm(ad)))
    return np.rad2deg(theta)


def _shock_angle(spec, n, field, leq90):
    if field.lower() == "b":
        a = spec["b_u"]
    else:
        a = spec["v_u"]

    theta = {}

    for fname in n:
        tmp = np.rad2deg(np.arccos(np.sum(a * n[fname]) / np.linalg.norm(a)))

        if tmp > 90.0 and leq90:
            theta[fname] = 90.0 - tmp
        else:
            theta[fname] = tmp

    return theta


def _shock_model(spec, *args):
    eps, l_, x_0, y_0, alpha = args

    # Rotation matrix
    rot_mat = R.from_euler("z", alpha, degrees=True).as_matrix()

    # offset from GSE
    r_0 = np.array([x_0, y_0, 0])

    # sc position in GSE (or GSM or whatever) in Earth radii
    if isinstance(spec["r_xyz"], xr.DataArray):
        # Time series
        r_sc = np.mean(spec["r_xyz"].data, axis=0) / 6371.0
    elif isinstance(spec["r_xyz"], (np.ndarray, list)) and len(spec["r_xyz"]) == 3:
        # Array like
        r_sc = spec["r_xyz"] / 6371.0
    else:
        raise TypeError("r_xyz must be a time series or a vector!!")

    def fval(sig, *args):
        rot_mat, r_sc, r_0, l_ = args

        # sc position in the natural system (cartesian)
        r_p = np.matmul(rot_mat, r_sc) - sig * r_0

        # sc polar angle in the natural system
        theta_p = np.arctan2(np.sqrt(r_p[1] ** 2 + r_p[2] ** 2), r_p[0])

        # minimize |LH-RH| in eq 10.22
        res = np.abs(sig * l_ / np.linalg.norm(r_p) - 1 - eps * np.cos(theta_p))
        return res

    # find the best fit for sigma
    sig_0 = optimize.minimize(fval, 1, args=(rot_mat, r_sc, r_0, l_))
    # to make sure it finds the largest sigma
    sig_0 = optimize.minimize(fval, 2 * sig_0.x[0], args=(rot_mat, r_sc, r_0, l_))

    # calculate normal
    r_p0 = np.matmul(rot_mat, r_sc) - sig_0.x[0] * r_0

    # gradient to model surface
    grad_s_x = (r_p0[0] * (1 - eps**2) + eps * sig_0.x[0] * l_) * np.cos(
        np.deg2rad(alpha)
    ) + r_p0[1] * np.sin(np.deg2rad(alpha))
    grad_s_y = -(r_p0[0] * (1 - eps**2) + eps * sig_0.x[0] * l_) * np.sin(
        np.deg2rad(alpha)
    ) + r_p0[1] * np.cos(np.deg2rad(alpha))
    grad_s_z = r_p0[2]

    grad_s = np.array([grad_s_x, grad_s_y, grad_s_z], dtype=np.float32)
    grad_s /= np.linalg.norm(r_p0, keepdims=True) * 2 * sig_0.x[0] * l_

    # normal vector
    n = grad_s / np.linalg.norm(grad_s, keepdims=True)

    return n, sig_0.x[0]


def _shock_speed(spec, n, theta_bn, method):
    if method.lower() == "gt" and "f_cp" in spec and "dt_f" in spec and "d2u" in spec:
        v_sh = _speed_gosling_thomsen(spec, n, theta_bn)
    elif method.lower() == "mf":
        v_sh = _speed_mass_flux(spec, n)
    elif method.lower() == "sb":
        v_sh = _speed_smith_burton(spec, n)
    elif method.lower() == "mo" and "f_cp" in spec and "dt_f" in spec and "d2u" in spec:
        v_sh = _speed_moses(spec, n, theta_bn)
    else:
        v_sh = {k: 0.0 for k in n}

    return v_sh


def _speed_gosling_thomsen(spec, n, theta_bn):
    v_sh = {}

    for k, nvec in n.items():
        theta = np.deg2rad(theta_bn[k])
        nvec = n[k]

        # Notation as in (Gosling and Thomsen 1985)
        w = 2 * np.pi * spec["f_cp"]
        t1 = np.arccos((1 - 2 * np.cos(theta) ** 2) / (2 * np.sin(theta) ** 2)) / w
        x_0 = w * t1 * (2 * np.cos(theta) ** 2 - 1) + 2 * np.sin(theta) ** 2 * np.sin(
            w * t1
        )
        x_0 /= w * spec["dt_f"]

        # the sign of Vsh in this method is ambiguous, assume n points upstream
        v_sh[k] = (
            spec["d2u"] * np.sum(spec["v_u"] * nvec) * (x_0 / (1 + spec["d2u"] * x_0))
        )

    return v_sh


def _speed_mass_flux(spec, n):
    rho_u = spec["n_u"] * constants.proton_mass
    rho_d = spec["n_d"] * constants.proton_mass

    v_sh = {}

    for k, nvec in n.items():
        v_sh[k] = (
            rho_d * np.sum(spec["v_d"] * nvec) - rho_u * np.sum(spec["v_u"] * nvec)
        ) / (rho_d - rho_u)

    return v_sh


def _speed_smith_burton(spec, n):
    v_sh = {}

    for k, nvec in n.items():
        v_sh[k] = np.sum(spec["v_u"] * nvec) + np.linalg.norm(
            np.cross(spec["delta_v"], spec["b_d"])
        ) / np.linalg.norm(spec["delta_b"])

    return v_sh


def _speed_moses(spec, n, theta_bn):
    v_sh = {}

    for k, nvec in n.items():
        theta = np.deg2rad(theta_bn[k])
        theta_vn = np.arccos(np.sum(nvec * spec["v_u"]) / np.linalg.norm(spec["v_u"]))

        # Notation as in (Moses et al., 1985)
        w = 2 * np.pi * spec["f_cp"]
        x = 0.68 * np.sin(theta) ** 2 * np.cos(theta_vn) / (w * spec["dt_f"])

        v_sh[k] = np.sum(spec["v_u"] * nvec) * (x / (1 + spec["d2u"] * x))

    return v_sh


def _constraint_matrix(spec, n):
    u_vecs = [
        spec["delta_b"],
        np.cross(spec["b_d"], spec["b_u"]),
        np.cross(spec["b_u"], spec["delta_v"]),
        np.cross(spec["b_d"], spec["delta_v"]),
        np.cross(spec["delta_b"], spec["delta_v"]),
    ]

    c_mat = np.zeros((len(u_vecs), len(n)))

    for i, u in enumerate(u_vecs):
        for j, nvec in enumerate(n.values()):
            c_mat[i, j] = np.sum(u * nvec) / np.linalg.norm(u)

    # fields that are by definition zero are set to 0
    c_mat[[0, 0, 0, 0, 1, 2, 2, 3, 3, 4, 4], [0, 2, 3, 4, 0, 1, 2, 1, 3, 1, 4]] = 0.0

    return c_mat



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def plot_heatmap(
    ax,
    data,
    row_labels,
    col_labels,
    cbar_kw: dict = None,
    cbarlabel: str = "",
    **kwargs,
):
    r"""Creates a heatmap from a numpy array and two lists of labels.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axis to which the heatmap is plotted.
    data : array_like
        A 2D numpy array of shape (N, M).
    row_labels : list or numpy.ndarray
        A list or array of length N with the labels for the rows.
    col_labels : list or numpy.ndarray
        A list or array of length M with the labels for the columns.
    cbar_kw : dict, Optional
        A dictionary with arguments to `matplotlib.Figure.colorbar`.
    cbarlabel : str, Optional
        The label for the colorbar.
    **kwargs
        All other arguments are forwarded to `imshow`.

    Returns
    -------
    im : matplotlib.image.AxesImage
        The AxesImage of the data.
    cbar : matplotlib.colorbar.Colorbar
        Colorbar.

    """

    # Plot the heatmap
    if cbar_kw is None:
        cbar_kw = {}

    im = ax.imshow(data, **kwargs)

    divider = make_axes_locatable(ax)

    colorbar_axes = divider.append_axes("right", size="2%", pad=0.1)

    # Create colorbar
    cbar = ax.figure.colorbar(im, ax=ax, cax=colorbar_axes, **cbar_kw)
    cbar.ax.set_ylabel(cbarlabel)

    # We want to show all ticks...
    ax.set_xticks(np.arange(data.shape[1]))
    ax.set_yticks(np.arange(data.shape[0]))
    # ... and label them with the respective list entries.
    ax.set_xticklabels(col_labels)
    ax.set_yticklabels(row_labels)

    # Let the horizontal axes labeling appear on top.
    ax.tick_params(top=True, bottom=True, labeltop=True, labelbottom=False)

    # Turn spines off and create white grid.
    # ax.spines[:].set_visible(False)

    ax.set_xticks(np.arange(data.shape[1] + 1) - 0.5, minor=True)
    ax.set_yticks(np.arange(data.shape[0] + 1) - 0.5, minor=True)
    ax.grid(which="minor", color="k", linestyle="-", linewidth=0.5)
    ax.tick_params(which="minor", bottom=False, left=False)
    ax.set_axisbelow(False)
    cbar.ax.set_axisbelow(False)

    return im, cbar



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

from .avg_4sc import avg_4sc

# Local imports
from .c_4_grad import c_4_grad
from .gradient import gradient
from .ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def st_diff(r_mms, b_mms, lmn):
    r"""Computes velocity of the structure using spatio-temporal
    derivative method [13]_ [14]_ as

    .. math::

        V_{str}^{LMN} = -d_t B^{LMN}\left [\nabla^{LMN} B^{LMN}\right]^T
        \left [ S^{LMN} \right ]^{-1}

    where :math:`B^{LMN}`, :math:`\nabla^{LMN}B^{LMN}`, :math:`S^{LMN}`
    and :math:`V_{str}^{LMN}` are namely the magnetic
    field, its gradient, its rotation rate tensor and the velocity of the
    structure in the LMN coordinates system.

    Parameters
    ----------
    r_mms : list of xarray.DataArray
        Spacecraft positions.
    b_mms : list of xarray.DataArray
        Background magnetic field.
    lmn : ndarray
        Structure coordinates system.

    Returns
    -------
    v_str : xarray.DataArray
        Velocity of the structure in its coordinates system.

    References
    ----------
    .. [13] Shi, Q. Q., Shen, C., Pu, Z. Y., Dunlop, M. W., Zong, Q. G.,
            Zhang, H., et al. (2005), Dimensional analysis of observed
            structures using multipoint magnetic field measurements:
            Application to Cluster. Geophysical Research Letters, 32,
            L12105. doi : https://doi.org/10.1029/2005GL022454.

    .. [14] Shi, Q. Q., Shen, C., Dunlop, M. W., Pu, Z. Y., Zong, Q. G.,
            Liu, Z. X., et al. (2006), Motion of observed structures
            calculated from multi‐point magnetic field measurements:
            Application to Cluster. Geophysical Research Letters, 33,
            L08109. doi : https://doi.org/10.1029/2005GL025073.
    """

    # Compute magnetic field at the center of mass of the tetrahedron
    b_xyz = avg_4sc(b_mms)

    # Gradient of the magnetic field
    grad_b = c_4_grad(r_mms, b_mms)

    # Time derivative of the magnetic field at the center of mass of the
    # tetrahedron
    db_dt = gradient(b_xyz)

    # Transform gradient to LMN frame
    l_grad_b = np.matmul(grad_b.data, lmn[:, 0])
    m_grad_b = np.matmul(grad_b.data, lmn[:, 1])
    n_grad_b = np.matmul(grad_b.data, lmn[:, 2])

    # Compute velocity of the structure using MDD
    v_str = np.zeros(db_dt.shape)
    v_str[:, 0] = np.sum(db_dt * l_grad_b, axis=1)
    v_str[:, 0] /= np.linalg.norm(l_grad_b, axis=1) ** 2
    v_str[:, 1] = np.sum(db_dt * m_grad_b, axis=1)
    v_str[:, 1] /= np.linalg.norm(m_grad_b, axis=1) ** 2
    v_str[:, 2] = np.sum(db_dt * n_grad_b, axis=1)
    v_str[:, 2] /= np.linalg.norm(n_grad_b, axis=1) ** 2

    # To time series
    v_str_xyz = ts_vec_xyz(b_xyz.time.data, -v_str)

    return v_str_xyz



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def plot_line(axis, inp, **kwargs):
    r"""Line plot of time series.

    Parameters
    ----------
    axis : matplotlib.pyplot.subplotsaxes
        Axis
    inp : xarray.DataArray
        Time series to plot

    Other Parameters
    ----------------
    **kwargs
        Keyword arguments control the Line2D properties.

    Returns
    -------
    axs :
        Axes.

    """

    if axis is None:
        _, axis = plt.subplots(1)

    if len(inp.shape) == 3:
        data = np.reshape(
            inp.data,
            (inp.shape[0], inp.shape[1] * inp.shape[2]),
        )
    else:
        data = inp.data

    time = inp.time
    axis.plot(time, data, **kwargs)

    if time.dtype == "<M8[ns]":
        locator = mdates.AutoDateLocator(minticks=3, maxticks=7)
        formatter = mdates.ConciseDateFormatter(locator)
        axis.xaxis.set_major_locator(locator)
        axis.xaxis.set_major_formatter(formatter)

    axis.grid(True, which="major", linestyle="-", linewidth="0.2", c="0.5")
    axis.yaxis.set_major_locator(mticker.MaxNLocator(4))

    return axis



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .feeps_omni import feeps_omni
from .feeps_remove_bad_data import feeps_remove_bad_data
from .feeps_remove_sun import feeps_remove_sun
from .feeps_spin_avg import feeps_spin_avg
from .feeps_split_integral_ch import feeps_split_integral_ch

# Local imports
from .get_feeps_alleyes import get_feeps_alleyes

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def get_feeps_omni(
    tar_var,
    tint,
    mms_id,
    verbose: bool = True,
    data_path: str = "",
    spin_avg: bool = False,
):
    r"""Computes the omni-directional energy spectrum of the target data unit
    for the target specie over the target energy range. The data are washed,
    splitted and sunlight contamination free.

    Parameters
    ----------
    tar_var : str
        Key of the target variable like
        {data_unit}_{specie}_{data_rate}_{data_lvl}.
    tint : list of str
        Time interval.
    mms_id : int or str
        Index of the spacecraft.
    verbose : bool, Optional
        Set to True to follow the loading. Default is True.
    data_path : str, Optional
        Path of MMS data. Default uses `pyrfu.mms.mms_config.py`
    spin_avg : bool, Optional
        Spin average the omni-directional flux. Default is False.

    Returns
    -------
    flux_omni : xarray.DataArray
        Energy spectrum of the target data unit for the target specie in omni
        direction.

    See Also
    --------
    pyrfu.mms.get_feeps_alleyes, pyrfu.mms.feeps_remove_bad_data,
    pyrfu.mms.feeps_split_integral_ch, pyrfu.mms.feeps_remove_sun,
    pyrfu.mms.feeps_omni, pyrfu.mms.feeps_spin_avg

    """

    # Get all telescopes
    dataset_feeps = get_feeps_alleyes(
        tar_var,
        tint,
        mms_id,
        verbose,
        data_path,
    )

    # Remove bad eyes and bad energy channels (lowest)
    dataset_feeps_washed = feeps_remove_bad_data(dataset_feeps)

    # Separate last channel
    dataset_feeps_clean, _ = feeps_split_integral_ch(dataset_feeps_washed)

    # Remove sunlight contamination
    dataset_feeps_clean_sun_removed = feeps_remove_sun(dataset_feeps_clean)

    # Compute omni.directional spectrum
    spec_feeps_omni = feeps_omni(dataset_feeps_clean_sun_removed)

    if spin_avg:
        spec_feeps_omni = feeps_spin_avg(
            spec_feeps_omni,
            dataset_feeps.spinsectnum,
        )

    return spec_feeps_omni



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Third party imports
import numpy as np

# Built-in imports
import tqdm
import xarray as xr
from scipy.constants import electron_mass, electron_volt, proton_mass, speed_of_light

# Local imports
from ..pyrf.datetime642iso8601 import datetime642iso8601
from ..pyrf.int_sph_dist import int_sph_dist
from ..pyrf.resample import resample
from ..pyrf.time_clip import time_clip
from ..pyrf.ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def reduce(vdf, xyz, dim: str = "1d", base: str = "pol", **kwargs):
    r"""Reduces (integrates) 3D distribution to 1D (line) or 2D (plane).
    Draft do not use!!

    Parameters
    ----------
    vdf : xarray.Dataset
        3D skymap velocity distribution function.
    xyz : xarray.DataArray or numpy.ndarray
        Transformation matrix from instrument frame to desired frame.
    base : str, Optional
        Base for the 2D projection either cartesian 'cart' (default) or
        polar 'pol'.
    **kwargs
        Keyword arguments

    Returns
    -------
    out : xarray.DataArray
        Time series of reduced velocity distribution function.

    """

    # Make sure the projection dimension amd base are correct
    assert dim.lower() in ["1d", "2d", "3d"], "Invalid projection dimension!!"
    assert base.lower() in ["cart", "pol"], "Invalid projection base!!"

    # Lower energy bound of instrument bins
    delta_energy_minu = xr.DataArray(
        vdf.attrs["delta_energy_minus"],
        coords=[vdf.time.data, vdf.idx0.data],
        dims=["time", "idx0"],
    )

    # Clip the distribution. If no time interval provided use the entire
    # time series.
    vdf_time = vdf.time.copy()
    tint = kwargs.get("tint", list(datetime642iso8601(vdf_time.data[[0, -1]])))
    vdf_time = time_clip(vdf_time, tint).copy()
    vdf_energy = time_clip(vdf.energy, tint).copy()
    delta_energy_minu = time_clip(delta_energy_minu, tint)
    vdf_phi = time_clip(vdf.phi, tint).copy()
    vdf_theta = vdf.theta.copy()
    vdf_data = time_clip(vdf.data, tint).copy()

    # make input distribution to SI units, s^3/m^6
    if vdf.data.attrs["UNITS"].lower() == "s^3/cm^6":
        vdf_data *= 1e12
    elif vdf.data.attrs["UNITS"].lower() == "s^3/m^6":
        vdf_data *= 1e0
    elif vdf.data.attrs["UNITS"].lower() == "s^3/km^6":
        vdf_data *= 1e-18
    else:
        raise ValueError("Invalid units!!")

    # Get VDF dimension (time, energy, phi, theta)
    # n_t, _, n_ph, _ = vdf_data.shape
    n_t, n_en, _, _ = vdf_data.shape

    # Construct or resample the time series of transformation matrix
    if isinstance(xyz, xr.DataArray):
        # If time series, clip to time interval and resample to VDF's time line
        xyz = time_clip(xyz, tint)
        xyz = resample(xyz, vdf_time).data
    elif isinstance(xyz, np.ndarray) and xyz.ndim == 2:
        assert xyz.shape == (3, 3), "xyz must be a transformation matrix!!"
        # If matrix, tile to VDF timeline
        xyz = np.tile(xyz, (n_t, 1, 1))
    else:
        raise TypeError("Invalid type for xyz")

    # Check that the transformation matrices are orthonormal direct
    x_phat_ts, y_phat_ts, z_phat_ts = [xyz[:, :, i] for i in range(3)]
    x_phat_ts /= np.linalg.norm(x_phat_ts, axis=1, keepdims=True)
    y_phat_ts /= np.linalg.norm(y_phat_ts, axis=1, keepdims=True)
    z_phat_ts = np.cross(x_phat_ts, y_phat_ts)
    z_phat_ts /= np.linalg.norm(z_phat_ts, axis=1, keepdims=True)
    y_phat_ts = np.cross(z_phat_ts, x_phat_ts)

    # Construct the time series of transformation matrix
    xyz_ts = np.transpose(
        np.stack([x_phat_ts, y_phat_ts, z_phat_ts]),
        [1, 2, 0],
    )

    # Set azimuthal angle projection grid
    # d_phi = kwargs.get("d_phi", 2 * np.pi / n_ph)
    # phi_grid = np.linspace(0, 2 * np.pi, n_ph) + d_phi / 2  # centers

    # Number of Monte Carlo iterations and weights
    n_mc = kwargs.get("n_mc", 100)
    weight = kwargs.get("weight", None)

    v_lim = kwargs.get("v_lim", [-np.inf, np.inf])  # Velocity grid limits
    a_lim = kwargs.get("a_lim", [-180.0, 180.0])  # Azimuthal angle limits

    # Spacecraft potential to account for photoelectrons. If not provided set
    # to 0 V.
    sc_pot = kwargs.get("sc_pot", ts_scalar(vdf_time.data, np.zeros(n_t)))
    sc_pot = resample(sc_pot, vdf_time).data

    # Threshold energy (e.g., to remove background noise). If not provided set
    # to 0 eV.
    lower_e_lim = kwargs.get("lower_e_lim", 0.0)

    if isinstance(lower_e_lim, xr.DataArray):
        # If time series, clip to time interval and resample to VDF's time line
        lower_e_lim = resample(lower_e_lim, vdf_time).data
    elif isinstance(lower_e_lim, float):
        # If float, tile to VDF's timeline
        lower_e_lim = np.tile(lower_e_lim, n_t)
    else:
        raise TypeError("Invalid lower_e_lim!!")

    # Set particle specie mass from VDF's attribute "species"
    if vdf.species.lower() == "electrons":
        m_p = electron_mass
    elif vdf.species.lower() == "ions":
        m_p = proton_mass
    else:
        raise ValueError("Invalid species!!")

    # Convert maximum energy from instrument to velocity (relativistically
    # correct)
    e_max = vdf_energy.data[0, -1] + vdf.attrs["delta_energy_plus"][0, -1]
    gamma_max = 1 + electron_volt * e_max / (m_p * speed_of_light**2)
    v_max = speed_of_light * np.sqrt(1 - 1 / gamma_max**2)  # m/s

    speed_grid = kwargs.get("vg", None)  # TODO : check that for no input!!
    speed_grid_edges = kwargs.get("vg_edges", None)

    # initiate projected f
    if speed_grid_edges is not None:
        n_vg = len(speed_grid_edges) - 1
        speed_grid_cart = None
        speed_grid = speed_grid_edges[:-1] + 0.5 * np.diff(speed_grid_edges)
    elif speed_grid is not None:
        n_vg = len(speed_grid)
        speed_grid_cart = None
    else:
        n_vg = 2 * n_en
        speed_grid_cart = np.linspace(-v_max, v_max, n_vg, endpoint=True)

    n_pr = int(dim[0])
    f_g = np.zeros([n_t, *[n_vg] * n_pr])
    all_v = {f"v{chr(120 + i)}": np.zeros((n_t, n_vg)) for i in range(n_pr)}

    for i_t in tqdm.tqdm(range(n_t), ncols=60):  # display progress
        # 3d data matrix for time index
        f_3d = np.squeeze(vdf_data.data[i_t, ...])  # s^3/m^6
        f_3d = f_3d.astype(
            np.float64,
        ).copy()  # convert to C contiguous float64

        # Energies
        energy = vdf_energy.data[i_t, :]
        energy = energy.astype(np.float64)  # Convert to float64

        # Assign zeros to phase-space density values corresponding to energies
        # below the threshold energy or below the spacecraft potential if
        # provided.
        thresh_e = np.max([lower_e_lim[i_t], sc_pot[i_t]])
        e_min_idx = np.where(energy - delta_energy_minu[i_t, :] > thresh_e)[0][0]
        f_3d[:e_min_idx] = 0.0

        # Correct energy shift due to spacecraft potential
        energy -= sc_pot[i_t]
        energy[energy < 0] = 0.0

        # Convert energy to velocity (relativistically correct)
        gamma = 1 + electron_volt * energy / (m_p * speed_of_light**2)
        speed = speed_of_light * np.sqrt(1 - 1 / gamma**2)  # m/s

        # azimuthal angle
        phi = vdf_phi.data[i_t, :].astype(np.float64)  # in degrees
        phi = np.deg2rad(phi - 180.0)  # in radians

        # elevation angle
        theta = vdf_theta.data.astype(np.float64)  # in degrees
        theta = np.deg2rad(theta - 90.0)  # in radians

        # Set velocity projection grid.
        if speed_grid is None:
            if base == "pol":
                if dim == "1d":
                    speed_grid = np.hstack((-np.flip(speed), speed))
                else:
                    # speed_grid = speed
                    raise NotImplementedError(
                        "2d projection on polar grid is not ready yet!!",
                    )
            else:
                speed_grid = speed_grid_cart
        else:
            pass

        options = {
            "xyz": xyz_ts[i_t, ...],
            "n_mc": n_mc,
            "weight": weight,
            "v_lim": v_lim,
            "a_lim": a_lim,
            "projection_dim": dim,
            "projection_base": base,
            "speed_grid_edges": speed_grid_edges,
        }

        tmpst = int_sph_dist(f_3d, speed, phi, theta, speed_grid, **options)

        f_g[i_t, ...] = tmpst["f"]

        for i in range(n_pr):
            all_v[f"v{chr(120 + i)}"][i_t, :] = tmpst[f"v{chr(120 + i)}"] / 1e3  # km/s

    # Build output as a time series with dimensions:
    #   - (time x vx) for 1D reduced distribution
    #   - (time x vx x vy) for 2D reduced distribution
    coords = [
        vdf_time.data,
        *[all_v[f"v{chr(120 + i)}"][0, :] for i in range(n_pr)],
    ]
    dims = ["time", *[f"v{chr(120 + i)}" for i in range(n_pr)]]
    out = xr.DataArray(f_g, coords=coords, dims=dims)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

from ..pyrf.ts_skymap import ts_skymap

# Local imports
from .psd_rebin import psd_rebin

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def vdf_to_e64(vdf_e32):
    r"""Recompile data into 64 energy channels. Time resolution is halved.
    Only applies to skymap.

    Parameters
    ----------
    vdf_e32 : xarray.Dataset
        Time series of the particle distribution with 32 energy channels.

    Returns
    -------
    vdf_e64 : xarray.Dataset
        Time series of the particle distribution with 32 energy channels.

    """

    time_r, vdf_r, energy_r, phi_r = psd_rebin(
        vdf_e32,
        vdf_e32.phi,
        vdf_e32.attrs.get("energy0"),
        vdf_e32.attrs.get("energy1"),
        vdf_e32.attrs.get("esteptable"),
    )

    energy_r = np.tile(energy_r, (len(vdf_r), 1))

    # Data attributes
    data_attrs = vdf_e32.data.attrs

    # Coordinates attributes
    coords_attrs = {k: vdf_e32[k].attrs for k in ["time", "energy", "phi", "theta"]}

    # Global attributes
    glob_attrs = vdf_e32.attrs

    # update delta_energy
    if "delta_energy_plus" in glob_attrs and "delta_energy_minus" in glob_attrs:
        log_energy = np.log10(energy_r[0, :])
        log10_energy = np.diff(log_energy)
        log10_energy_plus = log_energy + 0.5 * np.hstack(
            [log10_energy, log10_energy[-1]],
        )
        log10_energy_minus = log_energy - 0.5 * np.hstack(
            [log10_energy[0], log10_energy],
        )

        energy_plus = 10**log10_energy_plus
        energy_minus = 10**log10_energy_minus
        delta_energy_plus = energy_plus - energy_r
        delta_energy_minus = abs(energy_minus - energy_r)
        delta_energy_plus[-1] = np.max(vdf_e32.attrs["delta_energy_minus"][:, -1])
        delta_energy_minus[0] = np.min(vdf_e32.attrs["delta_energy_minus"][:, 0])

        glob_attrs["delta_energy_plus"] = delta_energy_plus
        glob_attrs["delta_energy_minus"] = delta_energy_minus

    vdf_e64 = ts_skymap(
        time_r,
        vdf_r,
        energy_r,
        phi_r,
        vdf_e32.theta.data,
        energy0=energy_r[0, :],
        energy1=energy_r[0, :],
        esteptable=np.zeros(len(time_r)),
        attrs=data_attrs,
        coords_attrs=coords_attrs,
        glob_attrs=glob_attrs,
    )

    return vdf_e64



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import bisect
import logging

# 3rd party imports
import numpy as np
import xarray as xr
from scipy import interpolate

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def _guess_sampling_frequency(ref_time):
    r"""Compute sampling frequency of the time line."""

    n_data = len(ref_time)

    sfy1 = 1 / (ref_time[1] - ref_time[0])
    sfy = None
    not_found = True

    if n_data == 2:
        sfy = sfy1
        not_found = False

    cur, max_try = [2, 10]

    while not_found and cur < n_data and cur - 3 < max_try:
        sfy = 1 / (ref_time[cur] - ref_time[cur - 1])

        if np.absolute(sfy - sfy1) < sfy * 0.001:
            not_found = False

            sfy = (sfy + sfy1) / 2
            break

        sfy = sfy1
        cur += 1

    if not_found:
        raise RuntimeError(f"Cannot guess sampling frequency. Tried {max_try:d} times")

    return sfy


def _average(inp_time, inp_data, ref_time, thresh, dt2):
    r"""Resample inp_data to timeline of ref_time, using half-window of dt2.
    Points above std*tresh are excluded. thresh=0 turns off this option.
    """

    try:
        out_data = np.zeros([len(ref_time), *inp_data.shape[1:]])
    except IndexError:
        inp_data = inp_data[:, None]
        out_data = np.zeros((len(ref_time), inp_data.shape[1]))

    for i, ref_t in enumerate(ref_time):
        idx_l = bisect.bisect_left(inp_time, ref_t - dt2)
        idx_r = bisect.bisect_right(inp_time, ref_t + dt2)

        idx = np.arange(idx_l, idx_r)

        if idx.size == 0:
            out_data[i, ...] = np.nan
        else:
            if thresh:
                std_ = np.std(inp_data[idx, ...], axis=0)
                mean_ = np.mean(inp_data[idx, ...], axis=0)

                assert any(np.isnan(std_))

                for j, stdd in enumerate(std_):
                    if not np.isnan(stdd):
                        idx_r = bisect.bisect_right(
                            inp_data[idx, j + 1] - mean_[j],
                            thresh * stdd,
                        )
                        if idx_r:
                            out_data[i, j + 1] = np.mean(
                                inp_data[idx[idx_r], j + 1],
                                axis=0,
                            )
                        else:
                            out_data[i, j + 1] = np.nan
                    else:
                        out_data[i, ...] = np.nan

            else:
                out_data[i, ...] = np.mean(inp_data[idx, ...], axis=0)

    if out_data.ndim > 1 and out_data.shape[1] == 1:
        out_data = out_data[:, 0]

    return out_data


def _resample_dataarray(inp, ref, method, f_s, window, thresh):
    r"""Resample for time series (xarray.DataArray)"""

    flag_do = "check"

    if method:
        flag_do = "interpolation"

    if f_s is not None:
        sfy = f_s
    elif window is not None:
        sfy = 1 / window
    else:
        sfy = None

    inp_time = inp.time.data.view("i8") * 1e-9
    ref_time = ref.time.data.view("i8") * 1e-9

    if flag_do == "check":
        if len(ref_time) > 1:
            if not sfy:
                sfy = _guess_sampling_frequency(ref_time)

            if len(inp_time) / (inp_time[-1] - inp_time[0]) > 2 * sfy:
                flag_do = "average"
                logging.info("Using averages in resample")
            else:
                flag_do = "interpolation"
        else:
            flag_do = "interpolation"

    assert flag_do in ["average", "interpolation"]

    if flag_do == "average":
        assert not method, "cannot mix interpolation and averaging flags"

        if not sfy:
            sfy = _guess_sampling_frequency(ref_time)

        out_data = _average(inp_time, inp.data, ref_time, thresh, 0.5 / sfy)

    else:
        if not method:
            method = "linear"

        # If time series agree, no interpolation is necessary.
        if len(inp_time) == len(ref_time) and all(inp_time == ref_time):
            out_data = inp.data.copy()
            coord = [ref.coords["time"].data]

            if len(inp.coords) > 1:
                for k in list(inp.dims)[1:]:
                    coord.append(inp.coords[k].data)

            out = xr.DataArray(
                out_data,
                coords=coord,
                dims=inp.dims,
                attrs=inp.attrs,
            )

            return out

        tck = interpolate.interp1d(
            inp_time,
            inp.data,
            kind=method,
            axis=0,
            fill_value="extrapolate",
        )
        out_data = tck(ref_time)

    coord = [ref.coords["time"]]

    if len(inp.coords) > 1:
        for k in list(inp.dims)[1:]:
            coord.append(inp.coords[k].data)

    out = xr.DataArray(out_data, coords=coord, dims=inp.dims, attrs=inp.attrs)

    return out


def _resample_dataset(inp, ref, **kwargs):
    r"""Resample for VDFs (xarray.Dataset)"""
    # Find time dependent zVariables and resample
    tdepnd_zvars = list(filter(lambda x: "time" in inp[x].dims, inp))
    out_dict = {k: _resample_dataarray(inp[k], ref, **kwargs) for k in tdepnd_zvars}

    # Complete the dictionary with non-time dependent zVaraiables
    ndepnd_zvars = list(filter(lambda x: x not in tdepnd_zvars, inp))
    out_dict = {**out_dict, **{k: inp[k] for k in ndepnd_zvars}}

    # Find array_like attributes
    arr_attrs = filter(
        lambda x: isinstance(inp.attrs[x], np.ndarray),
        inp.attrs,
    )
    arr_attrs = list(arr_attrs)

    # Initialize attributes dictionary with non array_like attributes
    gen_attrs = filter(lambda x: x not in arr_attrs, inp.attrs)
    out_attrs = {k: inp.attrs[k] for k in list(gen_attrs)}

    for k in arr_attrs:
        attr = inp.attrs[k]

        # If array_like attributes have one dimension equal to time length
        # assume time dependent. One option would be move the time dependent
        # array_like attributes to time series to zVaraibles to avoid
        # confusion
        if attr.shape[0] == len(inp.time.data):
            coords = [np.arange(attr.shape[i + 1]) for i in range(attr.ndim - 1)]
            dims = [f"idx{i:d}" for i in range(attr.ndim - 1)]
            attr_ts = xr.DataArray(
                attr,
                coords=[inp.time.data, *coords],
                dims=["time", *dims],
            )
            out_attrs[k] = _resample_dataarray(attr_ts, ref, **kwargs).data
        else:
            out_attrs[k] = attr

    out_attrs = {k: out_attrs[k] for k in sorted(out_attrs)}

    # Make output Dataset
    out = xr.Dataset(out_dict, attrs=out_attrs)

    return out


def resample(
    inp,
    ref,
    method: str = "",
    f_s: float = None,
    window: int = None,
    thresh: float = 0,
):
    r"""Resample inp to the time line of ref. If sampling of X is more than two
    times higher than Y, we average X, otherwise we interpolate X.

    Parameters
    ----------
    inp : xarray.DataArray or xarray.Dataset
        Time series to resample.
    ref : xarray.DataArray
        Reference time line.
    method : str, Optional
        Method of interpolation "spline", "linear" etc.
        (default "linear") if method is given then interpolate
        independent of sampling.
    f_s : float, Optional
        Sampling frequency of the Y signal, 1/window.
    window : int or float or ndarray, Optional
        Length of the averaging window, 1/fsample.
    thresh : float, Optional
        Points above STD*THRESH are disregarded for averaging

    Returns
    -------
    out : xarray.DataArray
        Resampled input to the reference time line using the selected method.


    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2015-10-30T05:15:20.000", "2015-10-30T05:16:20.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field and electric field

    >>> b_xyz = mms.get_data("e_gse_fgm_srvy_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("e_gse_edp_fast_l2", tint, mms_id)

    Resample magnetic field to electric field sampling

    >>> b_xyz = pyrf.resample(b_xyz, e_xyz)

    """

    message = "Invalid input type. Input must be xarray.DataArary or xarray.Dataset"
    assert isinstance(inp, (xr.DataArray, xr.Dataset)), message

    options = {"method": method, "f_s": f_s, "window": window, "thresh": thresh}

    if isinstance(inp, xr.DataArray):
        out = _resample_dataarray(inp, ref, **options)
    else:
        out = _resample_dataset(inp, ref, **options)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from .calc_dt import calc_dt

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def autocorr(inp, maxlags: int = None, normed: bool = True):
    r"""Compute the autocorrelation function

    Parameters
    ----------
    inp : xarray.DataArray
        Input time series (scalar of vector).
    maxlags : int, Optional
        Maximum lag in number of points. Default is None (i.e., len(inp) - 1).
    normed : bool, Optional
        Flag to normalize the correlation.

    Returns
    -------
    out : xarray.DataArray
        Autocorrelation function

    """

    # Check input type
    assert isinstance(inp, xr.DataArray), "inp must be a xarray.DataArray"

    # Check input dimension (scalar or vector)
    assert inp.ndim < 3, "inp must be a scalar or a vector"

    if inp.ndim == 1:
        x = inp.data[:, None]
    else:
        x = inp.data

    n_t = len(inp)

    if maxlags is None:
        maxlags = n_t - 1

    if maxlags >= n_t or maxlags < 1:
        raise ValueError(f"maxlags must be None or strictly positive < {n_t:d}")

    lags = np.linspace(-float(maxlags), float(maxlags), 2 * maxlags + 1, dtype=int)
    lags = lags * calc_dt(inp)

    out_data = np.zeros((maxlags + 1, x.shape[1]))

    for i in range(x.shape[1]):
        correls = np.correlate(x[:, i], x[:, i], mode="full")

        if normed:
            correls /= np.sqrt(np.dot(x[:, i], x[:, i]) ** 2)

        correls = correls[n_t - 1 - maxlags : n_t + maxlags]
        out_data[:, i] = correls[lags >= 0]

    if inp.ndim == 1:
        out = xr.DataArray(
            np.squeeze(out_data),
            coords=[lags[lags >= 0]],
            dims=["lag"],
        )
    else:
        out = xr.DataArray(
            out_data,
            coords=[lags[lags >= 0], inp[inp.dims[1]].data],
            dims=["lag", inp.dims[1]],
        )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def mean_bins(inp0, inp1, bins: int = 10):
    r"""Computes mean of values of y corresponding to bins of x.

    Parameters
    ----------
    inp0 : xarray.DataArray
        Time series of the quantity of corresponding to the bins.
    inp1 : xarray.DataArray
        Time series of the quantity to compute the binned mean.
    bins : int, Optional
        Number of bins.

    Returns
    -------
    out : xarray.Dataset
        Dataset with :
            * bins : xarray.DataArray
                bin values of the x variable.
            * data : xarray.DataArray
                Mean values of y corresponding to each bin of x.
            * sigma : xarray.DataArray
                Standard deviation.

    Examples
    --------
    >>> import numpy
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft indices

    >>> mms_list = numpy.arange(1,5)

    Load magnetic field and electric field

    >>> r_mms, b_mms = [[] * 4 for _ in range(2)]
    >>> for mms_id in range(1, 5):
    >>> 	r_mms.append(mms.get_data("R_gse", tint, mms_id))
    >>> 	b_mms.append(mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id))
    >>>

    Compute current density, etc

    >>> j_xyz, _, b_xyz, _, _, _ = pyrf.c_4_j(r_mms, b_mms)

    Compute magnitude of B and J

    >>> b_mag = pyrf.norm(b_xyz)
    >>> j_mag = pyrf.norm(j_xyz)

    Mean value of J for 10 bins of B

    >>> m_b_j = pyrf.mean_bins(b_mag, j_mag)

    """

    assert isinstance(inp0, xr.DataArray), "inp0 must be xaray.DataArray"
    assert isinstance(inp1, xr.DataArray), "inp1 must be xaray.DataArray"

    assert inp0.ndim == 1, "inp0 must be a scalar"
    assert inp1.ndim == 1, "inp1 must be a scalar"

    x_sort = np.sort(inp0.data)
    x_edge = np.linspace(x_sort[0], x_sort[-1], bins + 1)

    y_avg, y_std = [np.zeros(bins), np.zeros(bins)]

    for i in range(bins):
        idx_left = inp0.data > x_edge[i]
        idx_right = inp0.data < x_edge[i + 1]

        y_bins = inp1.data[idx_left * idx_right]

        y_avg[i], y_std[i] = [np.mean(y_bins), np.std(y_bins)]

    bins = x_edge[:-1] + np.median(np.diff(x_edge)) / 2

    out_dict = {
        "data": (["bins"], y_avg),
        "sigma": (["bins"], y_std),
        "bins": bins,
    }

    out = xr.Dataset(out_dict)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

anodes_theta = np.array(
    [
        123.75000,
        101.25000,
        78.750000,
        56.250000,
        33.750000,
        11.250000,
        191.25000,
        213.75000,
        236.25000,
        258.75000,
        281.25000,
        303.75000,
        326.25000,
        348.75000,
        168.75000,
        146.25000,
    ],
)


def hpca_calc_anodes(inp, fov: list = None, method: str = "mean"):
    r"""Averages over anodes (or a given field of view) for HPCA ion data.

    Parameters
    ----------
    inp : xarray.DataArray
        Ion flux; [nt, npo16, ner63], looking direction
    fov : list of float, Optional
        Field of view, in angles, from 0-360. Default is [0., 360.].
    method : {"mean", "sum"}, Optional
        Method. Default is "mean".

    Returns
    -------
    out : xarray.DataArray
        HPCA ion flux averaged over the anodes within the selected field of
        view.

    """

    if fov is None:
        fov = [0.0, 360.0]

    assert method in ["mean", "sum"]

    times = inp.time.data
    energies = inp.ccomp.data

    cond_ = np.logical_and(anodes_theta >= fov[0], anodes_theta <= fov[1])
    anodes_in_fov = np.where(cond_)[0]

    if method == "mean":
        updated_spectra = inp.data[:, anodes_in_fov, :].mean(axis=1)
    else:
        updated_spectra = inp.data[:, anodes_in_fov, :].sum(axis=1)

    out = xr.DataArray(
        updated_spectra,
        coords=[times, energies],
        dims=["time", "energy"],
    )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def eis_pad_spinavg(inp, spin_nums):
    r"""Calculates spin-averaged Pitch-Angle Distribution (PAD) for the EIS
    instrument.

    Parameters
    ----------
    inp : xarray.DataArray
        Pitch-Angle Distribution.
    spin_nums : xarray.DataArray
        Spin #s associated with each measurement.

    Returns
    -------
    out : xarray.DataArray
        Spin-averaged PAD.

    See Also
    --------
    pyrfu.mms.get_eis_allt, pyrfu.mms.eis_pad

    """
    _, spin_starts = np.unique(spin_nums.data, return_index=True)

    spin_times = np.zeros(len(spin_starts), dtype="<M8[ns]")
    spin_sum_flux = np.zeros(
        (len(spin_starts), len(inp.theta.data), len(inp.energy.data)),
    )

    current_start = 0
    # loop through the spins
    for i, spin_start in enumerate(spin_starts):
        idx_ = np.where(spin_nums.data == spin_nums.data[spin_start])[0]
        spin_sum_flux[i, :, :] = np.nanmean(inp.data[idx_, :, :], axis=0)
        spin_times[i] = inp.time.data[current_start]
        current_start = spin_start + 1

    out = xr.DataArray(
        spin_sum_flux,
        coords=[spin_times, inp.theta.data, inp.energy.data],
        dims=["time", "theta", "energy"],
    )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import dates, gridspec
from matplotlib.widgets import Button, SpanSelector

# Local imports
from ..plot import plot_line
from .mva import mva
from .new_xyz import new_xyz
from .norm import norm
from .time_clip import time_clip

__author__ = "Atlas Silverhult"
__email__ = "atlas.silverhult.9977@student.uu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

__all__ = ["mva_gui"]


def mva_gui(inp):
    r"""GUI to interactively perform minimum variance analysis (MVA) on
    time series data by selecting the time interval to apply MVA on.
    The return of this function is a callback to the GUI object
    and class attributes like the minimum variance direction vector are accesable
    through this callback by the method get_minvar().


    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the quantity to load into GUI and perform MVA on.

    Returns
    -------
    mva_callback :
        Returns MvaGui object to access attributes. In order to keep
        GUI responsive and interactive, a reference to this object is needed.

    """

    mva_callback = MvaGui(inp)

    return mva_callback


class MvaGui:
    r"""Class to display and update GUI elements of MVA."""

    def __init__(self, b):
        # Time series data
        self.b = b
        self.t = self.b.time.data

        # Figure and GUI
        self.fig = self.init_fig(self.b)

        axreset = self.fig.add_axes([0.91, 0.66, 0.04, 0.03])
        self.resetbutton = Button(axreset, "Reset", hovercolor="0.75")
        self.resetbutton.on_clicked(self.reset_selection)

        self.span = SpanSelector(
            self.fig.axes[0],
            self.update_onselect,
            "horizontal",
            useblit=True,
            props={"alpha": 0, "facecolor": "black"},
            interactive=True,
            minspan=0,
            drag_from_anywhere=True,
        )

        # Normal vec - to be calculated
        self.minvar = None
        self.errors = None

    @staticmethod
    def init_fig(b_xyz):
        r"""Initialize figure window with time series. Returns figure object
        to access the axes.
        """

        fig = plt.figure(figsize=(10, 8))
        legend_options = {
            "ncol": 1,
            "loc": "center left",
            "frameon": True,
            "framealpha": 1,
            "bbox_to_anchor": (1, 0.5),
        }
        gs = gridspec.GridSpec(
            3, 3, top=0.9, left=0.05, right=0.9, hspace=0.3, wspace=0.35
        )
        # Time series data
        ax1 = fig.add_subplot(gs[0, :])

        # MVA transformed frame
        fig.add_subplot(gs[1, :])

        # Min/Max hodogram
        fig.add_subplot(gs[2, 0])

        # Interm/Max hodogram
        fig.add_subplot(gs[2, 1])

        # Text display
        ax5 = fig.add_subplot(gs[2, 2])
        ax5.axis("off")

        # Plot b_xyz in the first axis
        plot_line(ax1, b_xyz)
        plot_line(ax1, norm(b_xyz), color="black")
        ax1.set_ylabel("$B$ [nT]")

        ax1.set_title("Time series for magnetic field")
        b_labels = ["$B_x$", "$B_y$", "$B_z$", "$|B|$"]
        ax1.legend(b_labels, **legend_options)
        return fig

    def reset_selection(self):
        r"""Resets MVA selection. Event is passed on from button click but has
        no meaning or use here.
        """

        ax3 = self.fig.axes[2]
        ax4 = self.fig.axes[3]
        ax5 = self.fig.axes[4]
        self.span.set_visible(False)
        self.update_fig(self.b)
        ax3.clear()
        ax4.clear()
        for txt in ax5.texts:
            txt.set_visible(False)
        plt.draw()

    def update_onselect(self, tmin, tmax):
        r"""This function is called when a span is selected with spanselector."""

        indmin, indmax = np.searchsorted(dates.date2num(self.t), (tmin, tmax))
        indmax = min(len(self.t) - 1, indmax)

        region_t = self.t[indmin:indmax]

        # convert times to datetime64
        tmin = np.datetime64(dates.num2date(tmin).replace(tzinfo=None))
        tmax = np.datetime64(dates.num2date(tmax).replace(tzinfo=None))

        b_xyz_clip = time_clip(self.b, [tmin, tmax])
        if len(region_t) >= 2:
            self.update_fig(b_xyz_clip)
            self.fig.canvas.draw_idle()

    @staticmethod
    def force_positive(mva_frame, b_xyz):
        r"""Force maximum variance direction to be positive."""

        frame = mva_frame[2]
        frame[:, 0] *= np.sign(max(frame[:, 0], key=abs))

        # Keep frame right-handed
        frame[:, 2] = np.cross(frame[:, 0], frame[:, 1])
        b_lmn = new_xyz(b_xyz, frame)
        return b_lmn, mva_frame[1], frame

    def update_fig(self, b_xyz_clip):
        r"""Update figure based on the clipped time series"""

        legend_options = {
            "ncol": 1,
            "loc": "center left",
            "frameon": True,
            "framealpha": 1,
            "bbox_to_anchor": (1, 0.5),
        }

        # Perform MVA on the clipped time series
        b_lmn_clip, lamb_clip, frame_clip = self.force_positive(
            mva(b_xyz_clip), b_xyz_clip
        )

        # Pick out time series for each component of the new magnetic field
        b_1 = b_lmn_clip[:, 0]
        b_2 = b_lmn_clip[:, 1]
        b_3 = b_lmn_clip[:, 2]

        # Pick out each eigenvector as the columns of frame_clip
        v1 = frame_clip[:, 0]
        v2 = frame_clip[:, 1]
        v3 = frame_clip[:, 2]

        # Define axse of figure
        ax2 = self.fig.axes[1]
        ax3 = self.fig.axes[2]
        ax4 = self.fig.axes[3]
        ax5 = self.fig.axes[4]

        # Clear axis graphs from previous selection
        ax2.clear()
        ax3.clear()
        ax4.clear()

        # Plot b in MVA frame given by frame_clip
        b_new = new_xyz(self.b, frame_clip)
        plot_line(ax2, b_new)
        plot_line(ax2, norm(b_new), color="black")
        b2_labels = ["max", "interm", "min", "abs"]
        ax2.legend(b2_labels, **legend_options)
        ax2.set_ylabel("$B$ [nT]")
        ax2.set_title("MVA frame")

        # Update B3/B1 hodogram
        ax3.axis("equal")
        ax3.plot(b_3, b_1, c="black")
        ax3.set_ylabel("max")
        ax3.set_xlabel("min")

        # Update B2/B1 hodogram
        ax4.axis("equal")
        ax4.plot(b_2, b_1, c="black")
        ax4.set_ylabel("max")
        ax4.set_xlabel("interm")

        # Text
        for txt in ax5.texts:
            txt.set_visible(False)

        val_textstring = (
            f"$\\lambda_1$ = {np.round(lamb_clip[0] ,2)}\n"
            f"$\\lambda_2$ = {np.round(lamb_clip[1] ,2)}\n"
            f"$\\lambda_3$ = {np.round(lamb_clip[2] ,2)}"
        )
        ratio_textstring = (
            f"\n"
            f"$\\lambda_1 / \\lambda_2$ ="
            f" {np.round(lamb_clip[0 ] /lamb_clip[1], 1)}\n"
            f"$\\lambda_2 / \\lambda_3$ = {np.round(lamb_clip[1 ] /lamb_clip[2], 1)}"
        )
        vec_textstring = (
            f"\n"
            f"$x_1$ = {np.round(v1 ,2)}\n"
            f"$x_2$ = {np.round(v2 ,2)}\n"
            f"$x_3$ = {np.round(v3 ,2)}"
        )
        fontsize = "large"
        ax5.text(
            0,
            1,
            val_textstring,
            fontsize=fontsize,
            ha="left",
            va="top",
            transform=ax5.transAxes,
        )
        ax5.text(
            0,
            0.5,
            ratio_textstring,
            fontsize=fontsize,
            ha="left",
            va="center",
            transform=ax5.transAxes,
        )
        ax5.text(
            0,
            0,
            vec_textstring,
            fontsize=fontsize,
            ha="left",
            va="center",
            transform=ax5.transAxes,
        )

        # Set minimum variance vector
        self.minvar = v3

        return b_lmn_clip, lamb_clip, frame_clip

    def get_minvar(self):
        r"""Access the minimum varience direction vector from mva_gui_class object
        via this method.
        """

        if self.minvar is not None:
            out = self.minvar
        else:
            raise RuntimeError("Normal has not been calculated yet")

        return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import logging

from ..pyrf.ts_append import ts_append
from .get_ts import get_ts

# Local imports
from .list_files import list_files

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def _tokenize(dataset_name):
    dataset = dataset_name.split("_")

    # Index of the MMS spacecraft
    probe = dataset[0][-1]

    var = {"inst": dataset[1], "tmmode": dataset[2], "lev": dataset[3]}

    try:
        var["dtype"] = dataset[4]
    except IndexError:
        pass

    return probe, var


def db_get_ts(
    dataset_name,
    cdf_name,
    tint,
    verbose: bool = True,
    data_path: str = "",
):
    r"""Get variable time series in the cdf file.

    Parameters
    ----------
    dataset_name : str
        Name of the dataset.
    cdf_name : str
        Name of the target field in cdf file.
    tint : array_like
        Time interval.
    verbose : bool, Optional
        Status monitoring. Default is verbose = True
    data_path : str, Optional
        Path of MMS data. Default uses `pyrfu.mms.mms_config.py`

    Returns
    -------
    out : xarray.DataArray
        Time series of the target variable.

    """

    probe, var = _tokenize(dataset_name)

    files = list_files(tint, probe, var, data_path=data_path)

    if verbose:
        logging.info("Loading %s...", cdf_name)

    out = None
    for file in files:
        out = ts_append(out, get_ts(file, cdf_name, tint))

    return out
